"""
<Program>
  node_transition_lib.py

<Started>
  July 29, 2009

<Author>
  Monzur Muhammad
  monzum@cs.washington.edu

<Purpose>
  This is a library file for the node transition scripts. It is a general 
  module so all the transition scripts will be able to import this file 
  and easily create a new transition script.

<Usage>
  Meant to be imported.

<Information>
  The transitional script is meant to seek out all the nodes that 
  are currently advertising and check their states. If necessary, 
  the state of the node is changed. Currently a node can be in three 
  different state: donated state, canonical state, and one percent state. 
  The donated state is when new nodes are discovered. This is usually 
  when a new user installs seattle and donates resources. The transitional
  scripts finds all the nodes that have been donated and processes them. 
  The information about the donated nodes are added to the database and 
  the node state is changed to the canonical state. The canonical state 
  is an important state, as this is the state where the nodes are ready to 
  be used, and be split up. All nodes should be able to come back to the 
  canonical state if anything goes wrong. At this point the canonical state 
  is changed to the onepercent state. This is where the 10% resource is
  taken and split into 10 sections (or into 1% resources). This 1% resource 
  is the vessel. Thus a node manager is responsible for 10 vessels (by 
  default, if only 10% resource is donated). In the future we might have 
  other states such as a twopercent state, in case we want to split the 
  resource differently.
"""

import os
import sys
import time
import random
import traceback

import django.db

from seattle import runonce
from repyportability import *

import seattlegeni.common.util.log

from seattlegeni.common.api import maindb
from seattlegeni.common.api import backend
from seattlegeni.common.api import lockserver
from seattlegeni.common.api import nodemanager

# For setting the backend authcode.
import seattlegeni.backend.config

from seattlegeni.common.util.decorators import log_function_call

from seattlegeni.common.exceptions import *

from seattlegeni.website import settings

# Import all the repy files.
add_dy_support(locals())
dy_import_module_symbols('advertise.r2py')
dy_import_module_symbols('rsa.r2py')
dy_import_module_symbols('listops.r2py')
dy_import_module_symbols('parallelize.r2py')
dy_import_module_symbols('random.r2py')





# Whether _init_node_transition_lib() has been called yet.
is_initialized = False





def _retrieve_state_key(key_file_name):
  """ Retrieve pubkey from file and return the dictionary form of key"""
  return rsa_file_to_publickey(os.path.join(settings.SEATTLECLEARINGHOUSE_STATE_KEYS_DIR, key_file_name))  


known_transition_states = ['canonical', 'acceptdonation', 
                           'movingto_onepercentmanyevents', 
                           'movingto_canonical', 'twopercent',
                           'onepercentmanyevents',
                           'movingto_twopercent']

transition_state_keys = {}

for state in known_transition_states:
  transition_state_keys[state] = _retrieve_state_key(state + '.publickey')




@log_function_call
def process_nodes_and_change_state(change_nodestate_function_tuplelist, transition_script_name, sleeptime, parallel_instances=1):
  """
  <Purpose>
    locate all the nodes in a certain state and apply the function thats in 
    the change_state_function_tuplelist and change the node state. Uses 
    public key to identify the nodes in different states.

  <Arguments>
    change_state_function_tuplelist - a list of tuples that contains all
      the info to change the state of a node.
      Tuple arguments:
        ((startstate_name, startstate_publickey), (endstate_name, endstate_publickey), 
         processfunction, errorfunction, processfunc_args)

    transition_script_name - a unique name used to get a process lock.

    sleeptime - the amount of time to sleep between node processing.

    parallelinstances - the number of threads to run in parallel.

  <Exceptions>
    None
  <Side Effects>
    None

  <Return>
    None
  """
  
  _init_node_transition_lib()
  
  log("Starting state transition script: "+transition_script_name+".....")

  # Get a processlock so multiple versions of the same script aren't running.
  if runonce.getprocesslock("process_lock."+transition_script_name) != True:
    log("The lock for "+transition_script_name+" is being held.\nExiting...\n\n")
    return

  # Continuously run the script and keep checking for nodes whose state needs to be changed.
  while True:
    do_one_processnode_run(change_nodestate_function_tuplelist, transition_script_name, parallel_instances)
    log("Finished running do_one_processnode_run(), going to sleep for "+str(sleeptime)+" seconds")  
    time.sleep(sleeptime)
    log("Waking up from sleep.....")





@log_function_call
def do_one_processnode_run(change_nodestate_function_tuplelist, transition_script_name, parallel_instances=1):
  """
  <Purpose>
    This is the actual script that runs and starts everything up. It gets the list 
    of nodes that needs to be changed. Then starts up the function that tranistions
    the nodes in parallel.

  <Arguments> 
    change_state_function_tuplelist - a list of tuples that contains all
      the info to change the state of a node.
      Tuple arguments:
        (startstate_name, endstate_name, processfunction, 
         errorfunction, mark_node_active, processfunc_args)

    transition_script_name - a unique name used to get a process lock.

    sleeptime - the amount of time to sleep in seconds between node processing.

    parallelinstances - the number of threads to run in parallel.

  <Exceptions>
    None
  <Side Effects>
    None

  <Return>
    [(success_processrun_count, fail_processrun_count)]

      success_processrun_count - how many nodes were processed sucessfully
      fail_processrun_count - how many nodes were processed unsuccessfully
  """
  
  result_list = []
  # Going through the list of node state functions.
  for change_nodestate_function_tuple in change_nodestate_function_tuplelist:
    # Get some of the information out of the nodestate_function tuple.
    startstate_name = change_nodestate_function_tuple[0]
    endstate_name = change_nodestate_function_tuple[1]

    startstate_publickey = transition_state_keys[startstate_name]



    # Lookup nodes with the startstate publickey, on failure log info and continue.
    log("Looking up nodes to transition from "+startstate_name+" state to "+endstate_name+" state")
    successful_lookup, nodeprocesslist = find_advertised_nodes(startstate_name, startstate_publickey)



    # If lookup was unsuccessful, then continue on. info is logged in find_advertised_nodes().
    if not successful_lookup:
      log("find_advertised_nodes() failed to look up nodes. Going on to the next nodestate_function_tuple")
      continue



    else:
      log("List of ip/NAT found in advertise_lookup(): " + str(nodeprocesslist))
      log("Length of nodeprocesslist: " + str(len(nodeprocesslist))) 
      log("Starting run_parallel_process to spawn threads to process nodes...")

    # Call the function that runs the processnode function on all the nodes in parallel.
    success_processrun_count, fail_processrun_count = run_parallel_processes(nodeprocesslist, 
                                                                             transition_script_name, 
                                                                             parallel_instances, 
                                                                             *change_nodestate_function_tuple)

   
    # Return the number of failures and successes.
    log("Succeeded on "+str(success_processrun_count)+" nodes, Failed on "+str(fail_processrun_count)+" nodes")
    result_list.append( (success_processrun_count, fail_processrun_count) )

    
    # Since we do a lot of lookup in the database we are going to have
    # a lot of database objects since django caches all the lookups.
    # We have to clear these queries so our memory usage doesn't grow
    # over time.
    django.db.reset_queries()

  return result_list   




@log_function_call
def run_parallel_processes(nodeprocesslist, lockname, parallel_instances, *processargs):
  """
  <Purpose>
    Create parallel handle and run the node process functions in parallel.

  <Arguments>
    nodeprocesslist - the list of all the nodes that needs to be processed

    lockname - the name of the lock that was acquired earlier. To ensure
      that the lock is still held    

    *processargs - this includes startstate, endstate, 
      nodeprocess_func, nodeerror_func and nodeprocess_args 

  <Exceptions>
    No exceptions are raised but on excepts, everything is logged.

  <Side Effects>
    None

  <Return>
    (success_count, failure_count)

      success_count - The number of nodes that we successfully changed state
      failure_count - The number of nodes that we failed to change state
  """

  try:
    # Note that processnode is a function.
    parallel_handle = parallelize_initfunction(nodeprocesslist, processnode, parallel_instances, *processargs)

  except: 
    log("Error: failed to set up parallelize_initfunction" + traceback.format_exc())
    return (0, len(nodeprocesslist)) 




  try:
    # Keep looping until all the threads in the parallel handle are finished running.
    while not parallelize_isfunctionfinished(parallel_handle):
      # Check to see if the process lock is still being held.
      if not runonce.stillhaveprocesslock("process_lock."+lockname):
        log("The lock for "+lockname+" has been lost. Exiting transitional script")
        sys.exit(1)
    
      # Sleep for a second to give the parallel_handle enought time to finish.
      time.sleep(1)

    # Get the results to the parallel_handle for the node processing.
    nodeprocessresults = parallelize_getresults(parallel_handle)


    success_count = 0
    failure_count = 0

    # Account for all the successful results.
    for nodename, returnvalue in nodeprocessresults['returned']:
      if returnvalue:
        log("Success on node "+nodename+". Returnvalue: "+str(returnvalue))
        success_count += 1
      else:
        log("Failure on node "+nodename+".")
        failure_count += 1


    # Write to the log for all the nodes that had an exception.
    for nodename, exceptionstring in nodeprocessresults['exception']:
      log("Failure on node "+nodename+"\nException String: "+exceptionstring+"\n")
      failure_count += 1

  
    # Write to the log for all the nodes that were aborted.
    for nodename in nodeprocessresults['aborted']:
      log("The processnode_function was aborted for the node "+nodename)
      failure_count += 1


  finally:
    # Clean up the parallel_handle and make sure that the handle is closed.
    log("Cleaning up and closing parallel_handle")
    parallelize_closefunction(parallel_handle)

  # Everything worked out fine in the parallel_handle.
  return (success_count, failure_count)




    
@log_function_call
def processnode(node_string, startstate_name, endstate_name, nodeprocess_func, nodeerror_func, mark_node_active, *nodeprocess_args):
  """
  <Purpose>
    First check the current state of the node, to ensure that it is in the
    correct state. Then run the nodeprocess_func on the node to process
    the node. Then set the new state of the node once the node has been
    processed properly.

  <Arguments>
    node_string - this is the node itself that is gotten from advertise_lookup,
      most likely an ip:port address or NAT:ip.

    startstate_name - the state in which the nodes are in now.

    endstate_name - the state we want to transition the nodes to.

    nodeprocess_func - the function that is used to process the node

    nodeerror_func - the function that is run, if there is an error

    mark_node_active - This bit determines if we want to mark the node
      as an active node or not.

    nodeprocess_args - the arguments for nodeprocess_func

  <Exceptions>
    NodeError - raised if the node is in a weird state, or if the 
      node does not exist

    NodemanagerCommunicationError - raised if problem communicating the node

    DatabaseError - raised if unable to access database

    UnexpectedError - if some unusual error occurs thats not caught by others

  <Side Effects>
    Database may get modified

  <Return>
    Return true if processing the node was successful, otherwise return false
    if any exceptions occured
  """
 
  # The pubkey for the state.
  startstate_pubkey = transition_state_keys[startstate_name]
  endstate_pubkey = transition_state_keys[endstate_name]

  # Make sure that the node is just not an empty string. 
  # The node_string could be bad due to bad advertise_lookup.
  if not node_string:
    raise NodeError("An empty node was passed down to processnode() with startstate: "+
                   startstate_name)

  # Note that the first portion of the node might be an ip or a NAT string.
  (ip_or_nat_string, port_num) = split_node_string(node_string)

  log("Starting to process node: "+node_string)

  # Try to retrieve the vessel dictionary for a node, 
  # on error raise a NodeError exception.
  try:
    node_info = nodemanager.get_node_info(ip_or_nat_string, port_num)
  except NodemanagerCommunicationError:
    raise
    

  # Extract the nodeID in order to acquire a lock
  nodeID =  _do_rsa_publickey_to_string(node_info['nodekey'])
  
  # Acquire a node lock
  lockserver_handle = acquire_node_lock(nodeID)

  try:
    log("Retrieving node vesseldict for node: "+node_string)   
    node_info = nodemanager.get_node_info(ip_or_nat_string, port_num)
   
    log("Successfully retrieved node_info for node: " + node_string)

    # If the nodes are in acceptdonationstate, update/check the database
    # to ensure that it matches the node information.
    if startstate_pubkey == transition_state_keys['acceptdonation']:
      add_new_node_to_db(node_string, node_info)         
      log("Successfully added node to the database for node" + node_string)
      log("The database should reflect the node information accurately")

    # Get the database object    
    database_nodeobject = maindb.get_node(nodeID)

    # Retrieve the node state and and the list of vessels.
    current_node_state_pubkey = get_node_state(node_info, database_nodeobject)
 
    # Make sure that the node is in the right state.
    if current_node_state_pubkey != startstate_pubkey:
      log("The node is not in the right transition state. NodeID is: " + nodeID +
          " Current state is " + str(current_node_state_pubkey) +
          ". Should be in state " + str(startstate_pubkey))
      raise NodeError("Node is no longer in the right state!")
  
    # Run the processnode function that was passed originally from the transition scripts.
    try:
      nodeprocess_func(node_string, node_info, database_nodeobject, *nodeprocess_args)
    except:
      log("Failed to process node: " + node_string)
      raise NodeError("Could not process node: " + node_string + traceback.format_exc())

    # Set the node state now that the node has been processed.
    if startstate_pubkey != endstate_pubkey:
      log("Trying to set new state for node: " + node_string)
      set_node_state(database_nodeobject, endstate_pubkey)
      
      # If the mark_node_active bit was set, then we want to mark the node
      # as active in the database. Until the node is marked active, the user
      # may not get credited. This is usually set at the final stage when all
      # vessels have been split.
      if mark_node_active:
        maindb.mark_node_as_active(database_nodeobject)
      else:
        maindb.mark_node_as_inactive(database_nodeobject)
        
      log("Finished setting new state " + endstate_name + " on node " + node_string) 

    else:
      log("Not setting node state: start state and end state are the same.")

  except NodeError:
    log("Node data problem when processing node: " + node_string + traceback.format_exc())
    return False    
  except NodemanagerCommunicationError:
    log("Node communication failed while processing node: " + node_string)
    return False
  except DatabaseError:
    log("Ran into problem accessing database while processing node: " + node_string)
    return False
  except UnexpectedError:
    log("Ran into some unexpected error while processing node: " + node_string)
    return False
  finally:
    release_node_lock(lockserver_handle, nodeID)

  #everything worked out fine
  return True  





@log_function_call
def get_node_state(node_info, database_nodeobject):
  """
  <Purpose>
    Given a node, find out what state the node is in (canonical, onepercent,
    acceptdonation). And return the state of the node. The node state should
    be held in the extra vessel, which should have only one userkey

  <Arguments>
    node_info - a dictionary that contains information about the node

    database_nodeobject - a node object that was retrieved from the
      database for the current node.

  <Exceptions>
    NodeError - raised if the extra vessel is not found in the node_info or
      if the extra_vessel has multiple state keys under 'userkeys'

  <Side Effects>
    None

  <Return>
    node_state - the state that the node is in.
  """

  log("Starting get node state for node " + database_nodeobject.node_identifier)
  extra_vessel_name = database_nodeobject.extra_vessel_name
  log("Name of extra vessel: " + extra_vessel_name)
 
  #check to see that the extra vessel exists in node_info
  if extra_vessel_name not in node_info["vessels"]:
    raise NodeError("The extra_vessel_name '" + extra_vessel_name + 
                    "' doesn't exist on the node. "+traceback.format_exc())

  extra_vessel_info = node_info["vessels"][extra_vessel_name]
  log("Extra vessel info: " + str(extra_vessel_info))

  #make sure that the extra vessel has only one state key
  if len(extra_vessel_info["userkeys"]) != 1:
    raise NodeError("The extra_vessel_name '" + extra_vessel_name +
                    "' doesn't contain one key, it contains " + 
                    str(len(extra_vessel_info["userkeys"])))
 
  return extra_vessel_info["userkeys"][0]

  

  
 
@log_function_call
def set_node_state (database_nodeobject, end_state):
  """
  <Purpose>
    given a list of vessels, change the userkeys of all the vessels.
    Only one of the vessel should carry the end_state, all other vessels
    will have empty state as transitional state. Keep in mind that the 
    userkey_list contains the transitional state of the node.
  
  <Arguments>
    end_state - the state that the node should end in.

    database_nodeobject - an object that has been retrieved from the database

  <Exceptions>
    NodeError - raised if unable to change node state
    
    NodemanagerCommunicationError - incase we are unable to retrieve node_info

  <Side Effects>
    None

  <Return>
    None
  """

  # Convert the end_state pubkey to string format and then set the key.
  final_state_str = _do_rsa_publickey_to_string(end_state)

  # Try to change the node state by setting the new state in the
  # userkeys list of the extra vessel
  try:
    backend.set_vessel_user_keylist(database_nodeobject, database_nodeobject.extra_vessel_name, [final_state_str])
  except:
    raise NodeError("Unable to change state of node: " + database_nodeobject.node_identifier + 
                    " "+traceback.format_exc())
  log("Successfully changed the node state to "+str(end_state))
  
    



@log_function_call
def add_new_node_to_db(node_string, node_info):
  """
  <Purpose> 
    when new nodes come online, they may be not have the right data in 
    the database. This function ensures that the database has the 
    appropriate data and the node_info data matches with the 
    database data. There are four different situations and we want
    to be in the fourth situation by the end of this function (if
    we don't run into any error along the way.)
    Situation1:
      Database has no entry for node or and does not have a 
      donation record. The node has the donor key as the 
      owner key.
    Situation2:
      Database has entry for the node but does not have any 
      donation record. The node has the donor key as the 
      owner key.
    Situation3:
      Database has entry for the node and it has a
      donation record. The node has the donor key as the 
      owner key.
    Situation 4:
      Database has entry for the node and it has a
      donation record. The node has the per-node key
      as the owner key.

  <Arguments>
    node_string - The node address in ip:port or NAT:port format

    node_info - A dictionary that contains information about the node

  <Exceptions>
    NodeError - if the node does not have a valid state

    DatabaseError - if there is any problem while creating node in database
      or creating record of donation in database or problem retrieving user
      object from the database

  <Side Effects>
    Database might get modified

  <Return>
    None
  """

  log("Node in acceptdonationpublickey. Calling add_newnode_to_db()" +
      " to check that node is correctly in database")
  
  log("Looking for donor key for node: " + node_string)
  log("Looking for the vessel that has the transition state in its 'userkeys'")

  nodeID = _do_rsa_publickey_to_string(node_info['nodekey'])

  (donated_vesselname, donor_key) = find_extra_vessel_name_and_donor_key(node_info, node_string)
    
  try:
    # Retrieve the node object if it's already in the database.
    # This is for the case if a database entry was created,
    # but the script ran into an error later on, causing the
    # state of the node to be not updated.
    database_nodeobject = maindb.get_node(nodeID)
  except DoesNotExistError:
    # If the node object does not exist in the database, add it.
    # This is situation 1, where nothing has been done yet.
    log("Database entry does not exist for node: " + node_string)
    database_nodeobject = create_new_node_object(node_string, node_info, donated_vesselname)
    log("Successfully added node to the database: " + node_string)    
     
  # We are finished adding node object, or retrieving node object from database.
  log("Retrieved node object successfully with nodeID: " + nodeID + 
      " with the node: " + str(database_nodeobject))
    

  # Check to see if the vessels owner key has been changed from the donor key
  # to the per node key. This is either situation 2 or situation 3.
  if donor_key != rsa_string_to_publickey(database_nodeobject.owner_pubkey):
    # Note that this is the case for if there is multiple vessels in 
    # the same node that are in the acceptdonation state. On the first 
    # run a database record was created for one vessel and a donation 
    # record was created. If we are looking at a second vessel that is
    # in the acceptdonation state, we want to raise an error and back out.
    # To do this we compare the vessel name of our original donated vessel
    # with the correct vesel name. If the names the same then everything is
    # fine as we are just recovering the vessel. If the vesselname is not
    # the same then this is a new vessel and we don't want to do anything.
    # Currently we only accept one donation from one node. In the future
    # this might change.
    if donated_vesselname != database_nodeobject.extra_vessel_name:
      raise NodeError("There is already a record of a vessel donated from node: " + node_string)
    
    log("Database has nodeID but node has donor key")
    log("Checking to make sure that the donation was credited")

    # Check to see if the donor was credited.
    donation_list = maindb.get_donations_from_node(database_nodeobject)

    # If the donation lenght is 0 then the user has not been credited
    # yet for their donation. This could be either due to this being
    # a new node that was just added or it could be because we are
    # recovering from a crash. This is in situation 2 if the length
    # of donation_list is 0
    if len(donation_list) == 0:
      log("The donation has not been credited yet for node " + node_string)
      create_donation_record(database_nodeobject, donor_key)
    else:
      log("The donation has already been credited for node " + node_string)

    log("Attempting to change the owner for vessel: " + donated_vesselname)
    
    # Convert the donation key to string format, and set the vessel owner key.
    # We should be now in situation 3 where the node exists in the database
    # and there is a donation record in the database.
    donation_owner_pubkey_str = _do_rsa_publickey_to_string(donor_key)
    backend.set_vessel_owner_key(database_nodeobject, donated_vesselname, 
                                 donation_owner_pubkey_str, 
                                 database_nodeobject.owner_pubkey)

    log("Successfully changed the owner_key to " + donation_owner_pubkey_str)

  return database_nodeobject
  # At this point we should be in situation 4 where everything is set 
  # properly. That is an entry for the node should exist in the database,
  # an entry for the donation should exist in the database and the 
  # per-node key should be the owner key for the node.





@log_function_call
def create_new_node_object(node_string, node_info, donated_vesselname):
  """
  <Purpose>
    Create a new node entry in the database with a newly generated
    per node key
  
  <Arguments>
    node_string - the address of the node 

    node_info - information about the node

    vesselname - the name of the extra vessel

  <Exceptions>
    None

  <Side Effects>
    None

  <Return>
    Returns a database node object
  """ 

  nodeID = _do_rsa_publickey_to_string(node_info['nodekey'])
  (ip_or_nat_string, port_num) = split_node_string(node_string)

  # Generate a new set of keys for the node. The pubkey gets passed
  # down to us while the private key is stored in a database elsewhere.
  log("Generating new keys for node owner_key....")   
  new_node_owner_pubkey = backend.generate_key(ip_or_nat_string+". "+nodeID)
  log("Generated publickey for node "+node_string+" : "+str(new_node_owner_pubkey))
  
  try:
    # Attempt to add new node to db.
    database_nodeobject = maindb.create_node(nodeID, ip_or_nat_string, port_num, 
                                                 node_info['version'], False, 
                                                 new_node_owner_pubkey, 
                                                 donated_vesselname) 
     
    log("Added node to the database with nodeID: " + nodeID)  
  except:
    raise DatabaseError("Failed to create node and add to database. " + traceback.format_exc())    
 
  return database_nodeobject




@log_function_call
def create_donation_record(database_nodeobject, donor_key):
  """
  <Purpose>
    To create a donation record for a user in the database

  <Arguments>
    database_nodeobject - a nodeobject that was retrieved from
      the database

  <Exceptions>
    DatabaseError - raised if problem accessing the database

  <Side Effects>
    Database will get modified

  <Return>
    None
  """

  # Retrieve the user object in order to give them credit
  # for their donation using the donor key
  try:
    database_userobject = maindb.get_donor(rsa_publickey_to_string(donor_key))
    log("Retrieved the userobject of the donor from database: " +
        str(database_userobject))
  except:
    raise DatabaseError("Failed to retrieve the userobject from the database. " + 
                        traceback.format_exc())

  donation_description = "Crediting user " + str(database_userobject)+ " for donation of node " + str(database_nodeobject)
      
  # Attempt to give the user credit for their donation
  try:
    log("Attempting to credit donation for the node " + database_nodeobject.node_identifier)
    maindb.create_donation(database_nodeobject, database_userobject, 
                           donation_description)
  except:
    raise DatabaseError("Failed to credit user for donation for user: " +
                        str(database_userobject) + traceback.format_exc())

  log("Registered the donation for the user with the donor_key: " + 
      str(donor_key))





@log_function_call
def find_extra_vessel_name_and_donor_key(node_info, node_string):
  """
  <Purpose>
    A function that is used to determine the name of the extra
    vessel and the the donor key that is associated with it.

  <Arguments>
    node_info - the information about the node

  <Exceptions>
    NodeError - raised if no vessel was found

  <Side Effects>
    None
 
  <Return>
    (donation_vesselname, donation_key) 
  """
  
  # Find the extra vessel of the node. Loop through all the vessels
  # until we find a vessel which has the acceptdonationpublickey in
  # its userkey's list. 
  for vesselname in node_info['vessels']:
    log("looking in vessel "+vesselname+".....")
    if transition_state_keys['acceptdonation'] in node_info['vessels'][vesselname]['userkeys']:
      donation_key = node_info['vessels'][vesselname]['ownerkey']
      donation_vesselname = vesselname
      log("Found donation key in vessel "+donation_vesselname+" : "+
         str(donation_key))
      return (donation_vesselname, donation_key)       

  raise NodeError("Node "+node_string+" has no vessel with the acceptdonation " + 
                  "state key in the userkeys.")





@log_function_call 
def find_advertised_nodes(startstate_name, startstate_publickey):
  """
  <Purpose>
    Looks up nodes with publickeys that matches startstate_publickey. 
    If advertise_lookup fails, it sleeps and tries again. Tries upto
    10 times before it returns.

  <Arguments>
    startstate_name - the name of the state. In order for logging
    startstate_publickey - the key with which the node is identified

  <Exception>
    Nothing is raised but on failure we log the failure

  <Side Effects>
    None

  <Return>
    success, nodelist
      success - boolean, states weather advertise lookup worked
      nodelist -list of nodes that was found. Empty if advertise lookup failed

  """
 
  log("Looking for nodes with public key: \n"+str(startstate_publickey))

  #lookup nodes with the startstate_publickey. Lookup upto 10MB of nodes
  #if advertise_lookup fails, try it 10 times before moving on
  advertise_lookup_fail_count = 0
  while True:
    try:
      nodeprocesslist = _do_advertise_lookup(startstate_publickey)
      return True, nodeprocesslist

    except:
      #increment fail count
      advertise_lookup_fail_count += 1
      log("advertise_lookup failed "+str(advertise_lookup_fail_count)+
          " times, while looking up nodes in "+startstate_name+ 
          " state with pub key:\n"+str(startstate_publickey))

      #if the lookup fails 10 times, break out of the while loop and log the info
      if advertise_lookup_fail_count >= 10:
        log("advertise_lookup failed 10 times...moving on to the next state transition function")
        return False, []

      log("Sleeping for 10 second before retrying advertise_lookup...")
      time.sleep(1)






@log_function_call
def split_vessels (node_string, node_info, database_nodeobject, resourcetemplate):
  """
  <Purpose>
    The purpose of this function is to take a node has an extra vessel that is
    big enough to be split into two or more vessels. The resources on a vessel
    is determined by the resourcetemplate that is provided as well as the usable
    ports on the node.

  <Arguments>
    node_string - the name of the node. ip:port or NAT:port

    node_info - a dictionary containing information about the node

    database_nodeobject - a database object for the node

    resourcetemplate - the file that has information about resources

  <Exceptions>
    NodeError - Error raised if node is not in the right state

    NodemanagerCommunicationError - raised if we cannot retrieve the usable ports for a node

    NodeProcessError - raised if unable to split vessels properly

    DatabaseError - raised if unable to modify the database

  <Side Effects>
    Database gets modified.

  <Return>
    None
  """

  log("Beginning to divide vessel on node: "+node_string)

  # Extract the ip/NAT and the port.
  # Note that the first portion of the node might be an ip or a NAT string.
  (ip_or_nat_string, port_num) = split_node_string(node_string)

  donated_vesselname = database_nodeobject.extra_vessel_name

  # Retrieve the usable ports list for the node and then shuffle
  # the ports so each vessel gets a random subset of the ports
  usable_ports_list = nodemanager.get_vessel_resources(ip_or_nat_string, port_num, donated_vesselname)['usableports']
  log("List of usable ports in node: "+node_string+". "+str(usable_ports_list))
  random.shuffle(usable_ports_list)

  #the vessel that we start with
  current_vessel = donated_vesselname
  log("Name of starting vessel: "+current_vessel)

  # Keep splittiing the vessel until we run out of resources.
  # Note that when split_vessel is called the left vessel
  # has the leftover (extra vessel)and the right vessel has
  # the vessel with the exact resources.
  while len(usable_ports_list) >= 10:
    desired_resourcedata = get_resource_data(resourcetemplate, usable_ports_list)

    #use the first 10 ports so remove them from the list of usable_ports_list
    used_ports_list = usable_ports_list[:10]
    usable_ports_list = usable_ports_list[10:]

    log("Ports we are going to use for the new vessel: "+str(used_ports_list))
    log("Starting to split vessel: "+current_vessel)

    # Split the current vessel. The exact vessel is the right vessel
    # and the extra vessel is the left vessel.
    try:
      leftover_vessel, new_vessel = backend.split_vessel(database_nodeobject, current_vessel, desired_resourcedata)
    except NodemanagerCommunicationError, e:
      # The object 'e' will already include traceback info that has the actual node error.
      # If the failure is due to inability to split further, that's ok.
      if 'Insufficient quantity:' in str(e):
        log("Could not split " + current_vessel + " any further due to insufficient resource/quantity. " + str(e))
        # We must break out of the while loop here. If we let the exception get,
        # raised, it will look like the transition failed.
        break
      raise

    log("Successfully split vessel: "+current_vessel+" into vessels: "+leftover_vessel+" and "+new_vessel)
    current_vessel = leftover_vessel

    # Make sure to update the database and record the new
    # name of the extra vessel as when backend.split_vessels()
    # is called, the old vessel does not exist anymore.
    # Instead two new vessels are created, where the first
    # vessel is the extra vessel with leftover resources
    # and the second vessel has the actual amount of resources
    maindb.set_node_extra_vessel_name(database_nodeobject, current_vessel)

    # Set the user_list for the new vesel to be empty. Remember that user_list is what determines
    # the transition state, and only the extra vessel should have this set.
    backend.set_vessel_user_keylist(database_nodeobject, new_vessel, [])
    log("Changed the userkeys for the vessel "+new_vessel+" to []")

    # Add the newly created vessel to the database and then add the ports associated with
    # the vessel to the database also.
    log("Creating a vessel record in the database for vessel "+new_vessel+" for node "+node_string)
    try:
      vessel_object = maindb.create_vessel(database_nodeobject, new_vessel)
      log("Setting the vessel ports in the database for vessel "+new_vessel+" with port list: "+str(used_ports_list))
      maindb.set_vessel_ports(vessel_object, used_ports_list)
    except:
      raise DatabaseError("Failed to create vessel entry or change vessel entry for vessel: " +
                          new_vessel + ". " + traceback.format_exc())


  log("Finished splitting vessels up for the node: "+node_string)





@log_function_call
def get_resource_data(resourcetemplate, usable_ports_list):
  """
  <Purpose>
    Create the resource_template and return it.

  <Arguments>
    resourcetemplate - the resource file

    usable_ports_list - the list of ports that the node has

  <Exception>
    None

  <Side Effects>
    None

  <Return>
    None
  """

  # Edit the resource file to add the resources in.
  resources_data = resourcetemplate % (str(usable_ports_list[0]), str(usable_ports_list[1]),
                                       str(usable_ports_list[2]), str(usable_ports_list[3]),
                                       str(usable_ports_list[4]), str(usable_ports_list[5]),
                                       str(usable_ports_list[6]), str(usable_ports_list[7]),
                                       str(usable_ports_list[8]), str(usable_ports_list[9]),
                                       str(usable_ports_list[0]), str(usable_ports_list[1]),
                                       str(usable_ports_list[2]), str(usable_ports_list[3]),
                                       str(usable_ports_list[4]), str(usable_ports_list[5]),
                                       str(usable_ports_list[6]), str(usable_ports_list[7]),
                                       str(usable_ports_list[8]), str(usable_ports_list[9]))

  return resources_data





@log_function_call
def combine_vessels(node_string, node_info, database_nodeobject):
  """
  <Purpose>
    The purpose of this function is to combine all the vessels of 
    a node into the extra vessel.

  <Arguments>
    node_string - the name of the node. ip:port or NAT:port

    node_info - the information about the node including the vesseldict

    database_nodeobject - This is the nodeobject that was retrieved from our database

    node_state_pubkey - This is the state that the node should be in. After all the 
      vessels are combined, the final vessel should have this as its state

  <Exceptions>
    NodeError - raised if unable to combine the vessels together

    DatabaseError - raised if unable to delete vessel records from database

  <Side Effects>
    None

  <Return>
    None
  """

  log("Beginning combine_vessels for the node: "+node_string)

  node_pubkey_string = database_nodeobject.owner_pubkey
    
  #This is the extra vessel or the vessel that has the transition state
  extra_vessel = database_nodeobject.extra_vessel_name 

  #the list that will hold all the vesselnames of the node
  vessel_list = get_vessel_list(node_info, node_pubkey_string, extra_vessel, node_string)

  # Combine all the vessels into one vessel.
  log("Trying to combine all the vessels...")
  create_combined_vessel(database_nodeobject, extra_vessel, vessel_list)

  try:
    # Delete all the vessel recoreds from database. 
    maindb.delete_all_vessels_of_node(database_nodeobject)
    log("Removed all the vessel records from the database for node: "+node_string)
  except:
    raise DatabaseError("Unable to delete all vessel records from the database for node "+
                        node_string+". " + traceback.format_exc())





@log_function_call
def create_combined_vessel(database_nodeobject, extra_vessel, vessel_list):
  """
  <Purpose>
    Combine all the vessels into one vessel, given a list of vessels

  <Arguments>
    database_nodeobject - a node object from the database

    extra_vessel - the name of the extra vessel

    vessel_list - a list of all the vessels that need to be
      combined with the extra vessel.

  <Exceptions>
    NodeError - raised if unable to combine vessels.

  <Side Effects>
    The individual vessels will not exist after this function is called

  <Return>
    A vessel name that is the result after all the vessels are 
    combined.
  """
  
  # Loop through all the vessels in the vessel list and do a 
  # join_vessels call to the backend to join each vessel with
  # the extra vessel. Be the end we should have only one vessel,
  # which should be the extra_vessel. Make sure to also update
  # the database with each iteration to ensure that the
  # extra_vessel_name gets updated appropriately.
  for current_vessel in vessel_list:
    try: 
      combined_vessel = backend.join_vessels(database_nodeobject, extra_vessel, current_vessel)
    except:
      raise NodeError("Unable to combine vessel " + current_vessel +
                      " with vessel " + extra_vessel + traceback.format_exc())
    
    # Update the database and change the name of the extra_vessel_name for the node
    maindb.set_node_extra_vessel_name(database_nodeobject, combined_vessel)

    log("Combined vessel " + extra_vessel + " and " + current_vessel + " into " + combined_vessel)
    log("Set database nodeobjects extra_vessel_name to " + database_nodeobject.extra_vessel_name)
    extra_vessel = combined_vessel

  log("Finished combining all the vessels for node: " + database_nodeobject.node_identifier)
  
  return extra_vessel





@log_function_call
def update_database(node_string, node_info, database_nodeobject, update_database_node):
  """
  <Purpose>
    The purpose of this function is to update the database

  <Arguments>
    node_string - the name of the node. ip:port or NAT:port

    node_info - a dictionary containing information about the node

    node_database - a node object from the database

    update_database_node - This is a function that updates the database. May be replaced
      in the future.

  <Exceptions>
    DatabaseError - raised if unable to modify the database

  <Side Effects>
    None

  <Return>
    None
  """

  log("Beginning update_database on node: "+node_string)

  # Extract the ip/NAT and the port.
  # Note that the first portion of the node might be an ip or a NAT string.
  ip_or_nat_string, port_num = split_node_string(node_string)

  # Update the database with the most up to date info about the node.
  # Currently only updates the node version
  log("Retrieved database node object for node: " + node_string)
  try:
    update_database_node(database_nodeobject, node_info, ip_or_nat_string, port_num)
  except:
    raise node_transition_lib.DatabaseError("Unable to update the database." + traceback.format_exc())

  log("updated node database record with version: "+str(database_nodeobject.last_known_version))

  return





@log_function_call
def update_database_node(database_nodeobject, node_info, ip_or_nat_string, port_num):
  """
  <Purpose>
    Update the database with the information provided

  <Arguments>
    database_nodeobject - a database object of the node

    node_info - information about the node including the vesseldict

    ip_or_nat_string - the ip or nat address of the node

    port_num - the port number that is being used for the port

    node_status - the status of the node that should be set

  <Exception>
    DatabaseError - raise if we could not update database

  <Side Effects>
    None

  <Return>
    database_object
  """

  version = node_info['version']

  # If the node objec was inactive, then we are going to make it active.
  if not database_nodeobject.is_active:
    log("The node "+ip_or_nat_string+":"+str(port_num)+" was inactive. Now activating")

  try:
    # Update all the database info about the node.
    maindb.record_node_communication_success(database_nodeobject, version, ip_or_nat_string, port_num)
  except:
    raise DatabaseError("Unable to modify database to update info on node: "+ip_or_nat_string)

  log("Updated the database for node: "+ip_or_nat_string+", with the latest info")

  return database_nodeobject





@log_function_call
def get_vessel_list(node_info, node_pubkey_string, extra_vessel, node_string):
  """
  <Purpose>
    Retrieve a list of vessels that are in this node that
    can be acquired. That is all the vessels except the 
    extra vessel.

  <Arguments>
    node_info - a dictionary with node information
    
    node_pubkey_string - this is the per-node key. Its the key
      that determines if we are the owner of the vessel.

    extra_vessel - the name of the extra vessel

  <Exceptions>
    None

  <Side Effects>
    None

  <Return>
    a list of all the vessel names in this node
  """

  vessel_list = []

  node_pubkey_dict = rsa_string_to_publickey(node_pubkey_string)

  # Go through all the vessels and check if we are the owner of the
  # vessel. If we are then we add the vessel to the vessel_list,
  # which are the list of vessels that needs to be combined. 
  log("Finding all the vessels in the node "+node_string+"...")
  for current_vessel in node_info['vessels']:
    # Check to see if the vessels belong to us (SeattleGENI).
    if node_info['vessels'][current_vessel]['ownerkey'] == node_pubkey_dict:
      # Add the vessel to the list.
      vessel_list.append(current_vessel)
      log("Added vessel "+current_vessel+" to vessel_list")
      
  # Remove the starting vessel from the vessel list.
  vessel_list.remove(extra_vessel)

  return vessel_list





def split_node_string(node_string):
  """
  Convert a node string which is the address of the node
  in the form of ip:port or NAT:port, into the ip and its
  port number and return the string version of the ip/NAT
  and the int form of the port number
  """

  ip_or_nat_string, port_string = node_string.split(":")
  port_num = int(port_string)
  return (ip_or_nat_string, port_num)





#helper function now for testing.
@log_function_call
def _do_advertise_lookup(startstate_publickey):
  """ Do an advertise lookup. This function is used mainly for easier
      testing purposes. """
  nodelist = advertise_lookup(startstate_publickey, maxvals = 10*1024*1024)
  # There can sometimes be empty strings in the list returned by advertise_lookup.
  while "" in nodelist:
    nodelist.remove("")
  return nodelist





def acquire_node_lock(nodeID):
  """ 
  Create a lockserver handle and acquire a lock
  and return the lockserver handle
  """
  # Initialize a lock.
  lockserver_handle = lockserver.create_lockserver_handle()
  log("Created lockserver_handle for use on node: "+nodeID)
  # Acquire a lock for the node.
  lockserver.lock_node(lockserver_handle, nodeID)
  log("Acquired node lock for nodeID: "+nodeID)
 
  return lockserver_handle





def release_node_lock(lockserver_handle, nodeID):
  """ Release the node lock and Destroy the lockserver handle"""
  #release the node lock and destroy lock handle
  lockserver.unlock_node(lockserver_handle, nodeID)
  log("released lock for node: "+nodeID)
  lockserver.destroy_lockserver_handle(lockserver_handle)
  log("Destroyed lockserver_handle for node: "+nodeID)
  




def noop(*args):
  """ in some cases I don't want to do anything (i.e. just change state)"""
  pass



def log(message):
  """
  <Purpose>
    log information with a time stamp
  
  <Arguments>
    message - the message to log

  <Exceptions>
    None

  <Side Effects>
    log file may get edited if sterr and stdout is redirected

  <Return>
    None
  """

  seattlegeni.common.util.log.info(message)
  





def _init_node_transition_lib():
  """
  <Purpose>
    Perform one-time initialization actions on startup. This will be taken
    care of automatically, scripts don't need to worry about calling this.
  <Arguments>
    None
  <Exceptions>
    TimeUpdateError
      If time update fails in init_nodemanager().
  <Side Effects>
    All necessary initalization is done.
  <Returns>
    None
  """

  global is_initialized
  
  if not is_initialized:
    
    is_initialized = True
    
    # Set the log level if we want to override the defaults.
    #seattlegeni.common.util.log.set_log_level(seattlegeni.common.util.log.LOG_LEVEL_DEBUG)

    # We don't need to maindb.init_maindb() because that happens automatically
    # when django creates a new database connection.
    
    nodemanager.init_nodemanager()

    backend.set_backend_authcode(seattlegeni.backend.config.authcode)
    




def _do_rsa_publickey_to_string(pubkey):
  """A helper function to retrieve string form of pubkey, used for testing."""
  return rsa_publickey_to_string(pubkey)



class UnexpectedError(Exception):
  """Exception if something unexpected goes wrong."""



class NodeProcessError(Exception):
  """This exception means the node is in an invalid state."""



class NodeError(Exception):
  """This exception means the node is in an invalid state."""



class DatabaseError(Exception):
  """This exception means that the database maindb had some kind of error or did not return properly"""



from django.conf.urls.defaults import *

# Uncomment the next two lines to enable the admin:
# from django.contrib import admin
# admin.autodiscover()

urlpatterns = patterns('seattlegeni.website.xmlrpc.dispatcher',
                       # top level urls and functions:
                       (r'', 'rpc_handler', {}, 'rpc_handler'),
                      )



"""
<Program>
  assertions.py

<Started>
  29 June 2009

<Author>
  Justin Samuel

<Purpose>
  This is a utility module to provide simple assert_* functions which throw a
  AssertionError if the assertion fails. This is mostly intended to provide
  simple type checking of arguments to functions where the programmer calling
  the function should have been sure to provide the correct type of data.
  
  Note: We define __all__ at the bottom of this script to make sure that we
  only export the assertions we've defined here, not anything else that may be
  imported or used in this module. Only names that start with "assert_" are
  exported.
  
  Note: AssertionError is a python built-in exception.
"""

from seattlegeni.website.control.models import Donation
from seattlegeni.website.control.models import GeniUser
from seattlegeni.website.control.models import Node
from seattlegeni.website.control.models import Vessel



def assert_str(value):
  """
  <Purpose>
    Ensure that value is a string (can be str or unicode).
  """
  # Check for basestring which is the superclass of str and unicode.
  if not isinstance(value, basestring):
    raise AssertionError("Expected string value, received " + str(type(value)))



def assert_str_or_none(value):
  """
  <Purpose>
    Ensure that value is a string (can be str or unicode) or None.
  """
  # Check for basestring which is the superclass of str and unicode.
  if not isinstance(value, basestring) and value is not None:
    raise AssertionError("Expected string value or None, received " + str(type(value)))



def assert_int(value):
  """
  <Purpose>
    Ensure that value is an integer (can be int or long).
  """
  if not isinstance(value, int) and not isinstance(value, long):
    raise AssertionError("Expected int or long value, received " + str(type(value)))



def assert_positive_int(value):
  """
  <Purpose>
    Ensure that value is an integer (can be int or long) and is greater than
    zero.
  """
  if not isinstance(value, int) and not isinstance(value, long):
    raise AssertionError("Expected int or long value, received " + str(type(value)))
  if value < 1:
    raise AssertionError("Expected a positive value, received: " + str(value))



def assert_bool(value):
  """
  <Purpose>
    Ensure that value is a boolean.
  """
  if not isinstance(value, bool):
    raise AssertionError("Expected bool value, received " + str(type(value)))



def assert_list(value):
  """
  <Purpose>
    Ensure that value is a list.
  """
  if not isinstance(value, list):
    raise AssertionError("Expected list value, received " + str(type(value)))



def assert_list_of_str(value):
  """
  <Purpose>
    Ensure that value is a list of strings.
  """
  assert_list(value)
  for item in value:
    assert_str(item)



def assert_donation(value):
  """
  <Purpose>
    Ensure that value is a Donation object.
  """
  if not isinstance(value, Donation):
    raise AssertionError("Expected Donation object, received " + str(type(value)))



def assert_geniuser(value):
  """
  <Purpose>
    Ensure that value is a GeniUser object.
  """
  if not isinstance(value, GeniUser):
    raise AssertionError("Expected GeniUser object, received " + str(type(value)))
  


def assert_node(value):
  """
  <Purpose>
    Ensure that value is a Node object.
  """
  if not isinstance(value, Node):
    raise AssertionError("Expected Node object, received " + str(type(value)))
  


def assert_vessel(value):
  """
  <Purpose>
    Ensure that value is a Vessel object.
  """
  if not isinstance(value, Vessel):
    raise AssertionError("Expected Vessel object, received " + str(type(value)))
  


# Many modules are using the line 'from seattlegeni.common.util.assertions import *'
# to import the assertions. We define __all__ so that we only export names from
# this module that start with the "assert_". This prevents us from exporting anything
# we happened to import as well as any non-private helper functions that may get
# defined. 
__all__ = []
for name in dir():
  if name.startswith("assert_"):
    __all__.append(name)




"""
   Author: Justin Samuel

   Start Date: 30 June 2009

   Description:

   This test starts the lockserver and uses multiple threads, each acting as a
   different lockserver client, to communicate with the lockserver. The requests
   result in contention for some locks and cause some of the requests to block.
   
   Each method of the lockserver is used at least once in a correct fashion (that
   is, this test does not test invalid requests).
"""

# Add to the path the directory that the lockserver module is in ('../').
# This assumes that the script will be run from the tests/ directory which
# is one directory below where the lockserver_daemon.py file is.
import sys
sys.path.append('..')

import lockserver_daemon as lockserver

import threading
import xmlrpclib
import time



LOCKSERVER_URL = "http://localhost:8010"

# How long to wait (in seconds) for the test to finish before giving up.
TIMEOUT = 5.0

# Minimum time to wait (in seconds) for any action that requires giving
# another thread a chance to do something.
SLEEP_TIME = 0.5

events = {}

events['first_client_first_request_finished'] = threading.Event()
events['second_client_about_to_make_first_request'] = threading.Event()
events['second_client_about_to_release_locks'] = threading.Event()
  
error_occurred = False


def report_error(message=None):
  global error_occurred
  error_occurred = True
  
  print message




class LockserverThread(threading.Thread):
  
  def run(self):
    lockserver.main()





class BaseClientThread(threading.Thread):
  
  proxy = xmlrpclib.ServerProxy(LOCKSERVER_URL)
  

  def run(self):
    try:
      # do_work() must be implemented by subclasses
      self.do_work()
      
    except Exception:
      report_error()
      raise




class FirstClientThread(BaseClientThread):
  
  def do_work(self):
    
    # This will be the first request made by any client, so it doesn't wait
    # for any event.
    print "First client started."
    
    session_id = self.proxy.StartSession()
    print "First client assigned session_id: " + session_id
    
    locks = {'user':['bob']}
    
    print "First client acquiring locks: " + str(locks)
    self.proxy.AcquireLocks(session_id, locks)
    print "First client acquired locks: " + str(locks)

    events["first_client_first_request_finished"].set()
    
    # Give the second and third clients a chance to make blocking requests
    # for the same lock.
    time.sleep(SLEEP_TIME * 3)
    
    print "First client releasing locks: " + str(locks)
    self.proxy.ReleaseLocks(session_id, locks)
    
    print "First client ending session"
    self.proxy.EndSession(session_id)
    
      



class SecondClientThread(BaseClientThread):
  
  def do_work(self):
    
    events["first_client_first_request_finished"].wait()
    
    print "Second client started."
    
    session_id = self.proxy.StartSession()
    print "Second client assigned session_id: " + session_id
    
    # The third client will wait for some time after this event is set
    # before it starts making lock requests.
    events["second_client_about_to_make_first_request"].set()

    # Request the same lock already held by the first client.
    locks = {'user':['bob']}
    
    print "Second client acquiring locks: " + str(locks)
    self.proxy.AcquireLocks(session_id, locks)
    print "Second client acquired locks: " + str(locks)
    
    # The above AcquireLocks call will block until
    # the lock is obtained.
    
    events["second_client_about_to_release_locks"].set()
    
    print "Second client releasing locks: " + str(locks)
    self.proxy.ReleaseLocks(session_id, locks)
    
    print "Second client ending session"
    self.proxy.EndSession(session_id)
    
      
      


class ThirdClientThread(BaseClientThread):
  
  def do_work(self):
    
    events["second_client_about_to_make_first_request"].wait()
    
    # Give the second client a chance to make its requests.
    time.sleep(SLEEP_TIME)

    print "Third client started."
    
    session_id = self.proxy.StartSession()
    print "Third client assigned session_id: " + session_id
    
    # Request the same lock already held by the first client and already being
    # waited for by the second client. Request multiple user locks just to make
    # it more interesting.
    locks = {'user':['bob', 'alice']}
    
    print "Third client acquiring locks: " + str(locks)
    self.proxy.AcquireLocks(session_id, locks)
    print "Third client acquired locks: " + str(locks)
    
    # The above AcquireLocks call will block until
    # the lock is obtained.
    
    # This event wait for the second client to release locks is to ensure that
    # the test will fail if the third client acquired its locks before the second
    # client. If that were to happen (that is, if the lock queue wasn't processed
    # in the correct order, for example), this event will never get set and
    # the test will timeout.
    events["second_client_about_to_release_locks"].wait()
    
    print "Third client releasing locks: " + str(locks)
    self.proxy.ReleaseLocks(session_id, locks)
    
    # Have this client acquire and release the same locks a few more times.
    print "Third client acquiring/releasing the following locks a few times and holding on to them in the end: " + str(locks)
    self.proxy.AcquireLocks(session_id, locks)
    self.proxy.ReleaseLocks(session_id, locks)
    self.proxy.AcquireLocks(session_id, locks)
    
    # Have this client acquire and release some node locks while they hold the user locks.
    nodelocks = {'node':['123', '456']}
    print "Third client acquiring/releasing the following locks a few times: " + str(nodelocks)
    self.proxy.AcquireLocks(session_id, nodelocks)
    self.proxy.ReleaseLocks(session_id, nodelocks)
    self.proxy.AcquireLocks(session_id, nodelocks)
    self.proxy.ReleaseLocks(session_id, nodelocks)
    
    print "Third client releasing locks: " + str(locks)
    self.proxy.ReleaseLocks(session_id, locks)
    
    print "Third client ending session"
    self.proxy.EndSession(session_id)
    
    print "Third client calling GetStatus()"
    status = self.proxy.GetStatus()
    
    expected_heldlockdict = {'node': {'123': {'queue': [], 'locked_by_session': None},
                                      '456': {'queue': [], 'locked_by_session': None}}, 
                             'user': {'bob': {'queue': [], 'locked_by_session': None},
                                      'alice': {'queue': [], 'locked_by_session': None}}}
    if status["heldlockdict"] != expected_heldlockdict:
      report_error("The heldlockdict returned from GetStatus() was not as expected: " + str(status["heldlockdict"]))
      
    expected_sessiondict = {}
    if status["sessiondict"] != expected_sessiondict:
      report_error("The sessiondict returned from GetStatus() was not as expected: " + str(status["sessiondict"]))





def fail_if_thread_alive_after_timeout(clientthread):
  
  clientthread.join(TIMEOUT)

  # If the thread is still alive after TIMEOUT seconds, it's blocked and
  # so the test failed.
  if clientthread.isAlive():
    print "A client thread is still alive after the TIMEOUT expired."
    print "Assuming the test failed (the client is probably blocked)."
    sys.exit(1)
    
    
    


def main():
  # Start the lockserver.
  LockserverThread().start()
  
  # Give the lockserver time to start.
  time.sleep(SLEEP_TIME)

  # Start the clients.
  firstclientthread = FirstClientThread()
  firstclientthread.start()
  
  secondclientthread = SecondClientThread()
  secondclientthread.start()
  
  thirdclientthread = ThirdClientThread()
  thirdclientthread.start()
  
  fail_if_thread_alive_after_timeout(firstclientthread)
  fail_if_thread_alive_after_timeout(secondclientthread)
  fail_if_thread_alive_after_timeout(thirdclientthread)
    
  if error_occurred:
    print "An error occurred. Test failed."
    sys.exit(1)
    
  # The lockserver thread is still alive at this point and will be killed when
  # the main thread exits.
  print "Test passed."
  sys.exit(0)
  
  

if __name__ == "__main__":
  main()



"""
<Program Name>
  models.py

<Started>
  July 15, 2009

<Author>
  Justin Samuel

<Purpose>
  This file contains definitions of model classes for the main database
  of SeattleGeni. These models are used by django to 1) create the database
  schema, and 2) to define an interface between django applications and the
  database.

  For information on how models are defined in django apps, see
  http://docs.djangoproject.com/en/dev/topics/db/models/

  Changes to these models should be done with corresponding changes
  to the design document at https://seattle.cs.washington.edu/wiki/SeattleGeniDesign
"""

import django

from django.db import models
from django.contrib.auth.models import User as DjangoUser

from seattlegeni.common.util import log





# First, we want to register a signal. This page recommends putting this code
# in models.py: http://docs.djangoproject.com/en/dev/topics/signals/

# Called when new database connections are created (see below).
def _prepare_newly_created_db_connection(sender, **kwargs):
  from seattlegeni.common.api import maindb
  maindb.init_maindb()

# If this is a modern-enough version of django to support specifying a function
# to be called on database connection creation, then have it call init_maindb()
# at that time. This is to help prevent init_maindb() from accidentally not
# being called when it should be.
if django.VERSION >= (1,1):
  # connection_created only exists with django >= 1.1
  import django.db.backends.signals
  django.db.backends.signals.connection_created.connect(_prepare_newly_created_db_connection)
else:
  log.error("You must use django >= 1.1 in order to support automatically " +
            "perform custom database connection initialization. (See settings.py)")






class GeniUser(DjangoUser):
  """
  Defines the GeniUser model. A GeniUser record represents a SeattleGeni user.
  
  By extending the DjangoUser model, django will still create a separate table
  in the database for the GeniUser model but will take care of making it look
  the same to us.
  """

  # The port which must be assigned to a vessel for the user to be able to
  # acquire that vessel.
  # Note: This field may go away later on if users are no longer always
  # assigned to the same port on all vessels. 
  usable_vessel_port = models.IntegerField("GeniUser's vessel port")

  # The text the user supplied which identifies the organization they are
  # affiliated with. 
  affiliation = models.CharField("Affiliation", max_length=200)
  
  # The user's public key which they use for communicating with nodes.
  # Note that the key is stored as a string "e n" where e and n are
  # decimal numbers. Because of this, max_length != max bits in the key.
  # We index this field with custom sql. See the file sql/geniuser.sql.
  user_pubkey = models.CharField("GeniUser's public key", max_length=2048)
  
  # The user's private key which they use for communicating with nodes.
  # This is not stored in the Key DB because the website needs access to it and
  # it is not used by us. The private key will only be stored when the the user
  # has had us generate their keys. The user will be encouraged through the
  # website to download this private key and have us delete our copy of it.
  # Note that the key is stored as a string "d p q" where d, p, and q are
  # decimal numbers. Because of this, max_length != max bits in the key.
  user_privkey = models.CharField("GeniUser's private key [!]", max_length=4096, null=True)
  
  # This is not a cryptographic key. This is an API key that we generate which
  # can be used by the user with the public SeattleGeni XML-RPC interface.
  # The purpose is to allow developers to use the XML-RPC interface without
  # requiring them to embed their passphrase in their source code. 
  api_key = models.CharField("API key", max_length=100, db_index=True)
  
  # The public key new donations use to indicate the donation is by this user.
  # The corresponding private key is always stored in the Key DB and is
  # accessible using this public key. The user never sees their donor keys.
  # Note that the key is stored as a string "e n" where e and n are
  # decimal numbers. Because of this, max_length != max bits in the key.
  # We index this field with custom sql. See the file sql/geniuser.sql.
  donor_pubkey = models.CharField("Donor public Key", max_length=2048)
  
  # The number of vessels the user is allowed to acquire regardless of whether
  # they have made donations. When the user makes donations, they still get
  # these free vessel credits, as well. This is stored in the database rather
  # than a constant for all users as there are some users we want to special
  # case and give extra free resources to. Unfortunately, "free" has more than
  # one meaning in English. This is more accurately "gratis_vessel_credits".
  free_vessel_credits = models.IntegerField("Free (gratis) vessel credits", db_index=True)
  
  # Have the database keep track of when each record was created and modified.
  date_created = models.DateTimeField("Date added to DB", auto_now_add=True, db_index=True)
  date_modified = models.DateTimeField("Date modified in DB", auto_now=True, db_index=True)
  
  
  def __unicode__(self):
    """
    Produce a string representation of the GeniUser instance.
    """
    return "GeniUser:%s" % (self.username)





class Node(models.Model):
  """
  Defines the Node model. A Node record represents an individual nodemanager.
  When a node goes offline, it is marked inactive. Node records are never
  deleted because if they were and a node with a delete node's node_identifier
  came back online, the owner private key would have been lost. 
  """

  # The node's identifier (which happens to be a public key with no
  # corresponding private key).
  # This should be unique, but we can't set this constraint at the
  # database-level because the length is too long for mysql to index the full
  # field.
  # We index this field with custom sql. See the file sql/node.sql.
  node_identifier = models.CharField("Node identifier", max_length=2048)

  # The IP address the nodemanager was last known to be accessible through.
  last_known_ip = models.CharField("Last known nodemanager IP address or NAT string", max_length=100, db_index=True)

  # The port the nodemanager was last known to be accessible through. 
  last_known_port = models.IntegerField("Last known nodemanager port", db_index=True)

  # The version of seattle the node was last known to be running. 
  last_known_version = models.CharField("Last known version", max_length=64, blank=True, db_index=True)

  # The last time the node could be contacted.
  # This is set to the current time when the node object is first created.
  date_last_contacted = models.DateTimeField("Last date successfully contacted", auto_now_add=True, db_index=True)

  # The node gets marked as not active when it becomes inaccessible. Nodes are
  # never deleted from the database no matter how long they have been inactive.
  # If they were deleted from the database, we wouldn't have the owner key to
  # be able to contact them if they came back online later.
  is_active = models.BooleanField(db_index=True)
  
  # The node gets marked as broken if we detect that the state of the node
  # doesn't exactly match what we have in our database. By 'state', we don't
  # just mean the state being advertised (the user key of the extra vessel),
  # but rather all aspects of the node that we have in our database which
  # aren't trivial to fix. This includes the extra vessel not having the node
  # state key, vessels that our database has for the node not actually existing
  # on the node, etc. A node that is broken generally implies some human
  # intervention will be required to repair things, as well as find out what
  # caused the breakage.
  is_broken = models.BooleanField(db_index=True)

  # The SeattleGeni's owner key for this node. The private key is always stored
  # in the Key DB and is accessible using this public key.
  # Note that the key is stored as a string "e n" where e and n are
  # decimal numbers. Because of this, max_length != max bits in the key.
  owner_pubkey = models.CharField("Owner public key", max_length=2048)

  # The extra vessel will (at least in the near future) have the node's free
  # resources assigned to it, so the name needs to be known in order to do
  # things with those resources. 
  extra_vessel_name = models.CharField("Extra-vessel name", max_length=8, db_index=True)
  
  # Have the database keep track of when each record was created and modified.
  date_created = models.DateTimeField("Date added to DB", auto_now_add=True, db_index=True)
  date_modified = models.DateTimeField("Date modified in DB", auto_now=True, db_index=True)


  def __unicode__(self):
    """
    Produces a string representation of the Node instance.
    """
    return "Node:%s:%s:%d" % (self.node_identifier[:10].replace(" ", "_"), self.last_known_ip, self.last_known_port)





class Donation(models.Model):
  """
  Defines the Donation model. A Donation record represents the resources a user
  has donated from a single node. 
  """
  
  class Meta:
    # Only one record can have a given node and donor combination.
    unique_together = ("node", "donor")


  # The node that this donation is from.
  node = models.ForeignKey(Node, db_index=True)
  
  # The user that is credited for this donation.
  donor = models.ForeignKey(GeniUser, db_index=True)
 
  # This field will be used, if necessary, to indicate steps in the process of
  # setting up a donation's resources for use (which, for now, would be
  # creating vessels on the corresponding node using this donation's resources).
  # TODO: remove this if it doesn't get used.
  status = models.CharField("Donation status", max_length=100, blank=True, db_index=True)

  # Simple storage of the contents of a resource file describing the donated
  # resources. This information is not currently used anywhere. How this data
  # is stored may change in the future and this field just serves to start
  # keeping track of the information for now. 
  resource_description_text = models.TextField("Resource description")
  
  # Have the database keep track of when each record was created and modified.
  date_created = models.DateTimeField("Date added to DB", auto_now_add=True, db_index=True)
  date_modified = models.DateTimeField("Date modified in DB", auto_now=True, db_index=True)
    
    
  def __unicode__(self):
    """
    Produces a string representation of the Donation instance.
    """
    return "Donation:[%s]:[%s]" % (self.node, self.donor)





class Vessel(models.Model):
  """
  Defines the Vessel model. A vessel record represents a vessel that
  SeattleGeni has setup on a node. Note that this is not tied to an
  individual donation of resources from that node. 
  """

  class Meta:
    # Only one record can have a given node and vessel name combination.
    unique_together = ("node", "name")


  # The node that this vessel is on.
  node = models.ForeignKey(Node, db_index=True)

  # The name used to refer to the vessel when communicating with the nodemanager.
  name = models.CharField("Vessel name", max_length=50, db_index=True)
  
  # If this vessel has been acquired by a user, this is the user who acquired
  # it. This will be a null/None value if the vessel is not currently acquired
  # by any user.
  acquired_by_user = models.ForeignKey(GeniUser, null=True, db_index=True)
  
  # The date that the acquired_by_user acquired this vessel.
  date_acquired = models.DateTimeField("Date acquired", null=True, db_index=True)
  
  # The date after which the vessel should be taken away from the user who has
  # acquired it.
  date_expires = models.DateTimeField("Date that acquisition expires", null=True, db_index=True)
    
  # The vessel is marked as dirty if it needs to be reset, etc. before it can
  # be acquired.
  is_dirty = models.BooleanField(db_index=True)
  
  # The vessel's user keys can change due to the user who acquired the vessel
  # changing their key or providing vessel access to other users. A value of
  # False here indicates that the user keys have changed since they were last
  # set on the vessel.
  user_keys_in_sync = models.BooleanField(db_index=True)
    
  # Have the database keep track of when each record was created and modified.
  date_created = models.DateTimeField("Date added to DB", auto_now_add=True, db_index=True)
  date_modified = models.DateTimeField("Date modified in DB", auto_now=True, db_index=True)
    
    
  def __unicode__(self):
    """
    Produces a string representation of the Vessel instance.
    """
    return "Vessel:[%s]:%s" % (self.node, self.name)





class VesselPort(models.Model):
  """
  Defines the VesselPort model. A VesselPort record represents a port that is
  assigned to a vessel. A single vessel can have multiple ports assigned to it
  (multiple VesselPort records). 
  """

  class Meta:
    # Only one record can have a given vessel and port combination.
    unique_together = ("vessel", "port")


  # The vessel that this port record is in reference to.
  vessel = models.ForeignKey(Vessel, db_index=True)
  
  # The port being represented.
  port = models.IntegerField("Port", db_index=True)


  def __unicode__(self):
    """
    Produces a string representation of the VesselPort instance.
    """
    return "VesselPort:[%s]:%s" % (self.vessel, self.port)





class VesselUserAccessMap(models.Model):
  """
  Defines the VesselUserAccessMap model. A VesselUserAccessMap record
  represents user access to vessels. This is a many-to-many relationship.
  The user who acquired the vessel will always have a mapping to that vessel.
  In the future when additional users can be added to a Vessel through
  SeattleGeni, the additional users would have records here. 
  """
  
  class Meta:
    # Only one record can have a given vessel and user combination.
    unique_together = ("vessel", "user")


  vessel = models.ForeignKey(Vessel, db_index=True)
  
  user = models.ForeignKey(GeniUser, db_index=True)
  
  # Have the database keep track of when each record was created.
  date_created = models.DateTimeField("Date added to DB", auto_now_add=True, db_index=True)
  
  
  def __unicode__(self):
    """
    Produces a string representation of the VesselUserAccessMap instance.
    """
    return "VesselUserAccessMap:[%s]:[%s]" % (self.vessel, self.user)





class ActionLogEvent(models.Model):
  """
  Defines the ActionLogEvent model. An ActionLogEvent record represents an
  action performed in the system, either by a user or by other components
  whose actions are logged.
  """
  # The name used to refer to the type of action.
  function_name = models.CharField("Function type", max_length=50, db_index=True)
  
  # A string representation of the second argument to the function if that
  # argument was not a vessel list.
  second_arg = models.CharField("Second arg", null=True, max_length=50)
  
  # A string representation of the third argument to the function if that
  # argument was not a vessel list.
  third_arg = models.CharField("Third arg", null=True, max_length=50)
  
  # If the owner_type is "user", this will be the user that performed the
  # action.
  user = models.ForeignKey(GeniUser, null=True, db_index=True)
  
  # Whether the action was successful.
  was_successful = models.BooleanField(db_index=True)
  
  # A status message, error message, or other relevant information.
  message = models.CharField("Message", null=True, max_length=1024)
  
  # The number of affected vessels. This is actually redundant as it is just
  # the number of ActionLogVesselDetails records for this event. However, this
  # makes some admin-side display code easier.
  vessel_count = models.IntegerField("Vessel count", null=True, db_index=True)
  
  # When the action was started.
  date_started = models.DateTimeField("Date started", db_index=True)
  
  # The number of seconds it took for the action to be completed.
  completion_time = models.FloatField("Completion time (seconds)", db_index=True)
  
  def __unicode__(self):
    """
    Produces a string representation of the ActionLogEvent instance.
    """
    return "ActionLogEvent:[%s]:[%s]" % (self.id, self.function_name)





class ActionLogVesselDetails(models.Model):
  """
  Defines the ActionLogVesselDetails model. An ActionLogVesselDetails record
  represents details of a particular vessel related to an ActionLogEvent. There
  can be multiple ActionLogVesselDetails for a single ActionLogEvent.
  """
  # The ActionLogEvent record that this details record corresponds to.
  event = models.ForeignKey(ActionLogEvent, db_index=True)

  # The node that this vessel is on.
  node = models.ForeignKey(Node, db_index=True)
  
  # To know the ip or nat string despite the fact that this may change for the
  # node record itself.
  node_address = models.CharField("Node address", max_length=100, db_index=True)

  # To know the port despite the fact that this may change for the
  # node record itself.
  node_port = models.IntegerField("Node port", max_length=100, db_index=True)

  # We store the vessel name rather than a foreign key to the vessels table
  # because vessels can be deleted from the database, whereas nodes can't.
  vessel_name = models.CharField("Vessel name", max_length=50, db_index=True)

  def __unicode__(self):
    """
    Produces a string representation of the ActionLogVesselDetails instance.
    """
    return "ActionLogVesselDetails:[%s]:[%s]:[%s]" % (self.event, self.node_address,
                                                      self.vessel_name)



# Warning: settings must be imported and the database values modified before
# anything else is imported. Failure to do this first will result in django
# trying to create a test mysql database.
from seattlegeni.website import settings

import os
import sys
import tempfile

# Use a database file that actually resides in memory, if possible.
# Note that we do this instead of use TEST_DATABASE_NAME = ":memory:" because
# there seems to be a problem with additional threads accessing the test
# database in memory when using an in-memory sqlite database (":memory:")
# rather than a file-backed one.
# The use of Python 2.5 with /dev/shm as the location of the sqlite db may also
# may be the cause of some tests failing on testbed-opensuse due to locks not
# being released, most likely, so we'll only use /dev/shm for 2.6+.
if os.path.exists("/dev/shm") and sys.version_info >= (2, 6):
  sqlite_database_file_dir = "/dev/shm"
else:
  sqlite_database_file_dir = None

# Do not change the DATABASE_ENGINE. We want to be sure that sqlite3 is used
# for tests.
settings.DATABASE_ENGINE = 'sqlite3'

# We just want the name of a temporary file that we'll use multiple times.
# We don't want it to exist already, so we remove it after creating it. This is
# not for anything security sensitive, it will only be used for test sqlite
# databases. We could just make up a random name of our, really.
test_db_filehandle, test_db_filename = tempfile.mkstemp(dir=sqlite_database_file_dir, suffix=".sqlite")
os.close(test_db_filehandle)
os.remove(test_db_filename)

settings.TEST_DATABASE_NAME = test_db_filename
# We set the DATABASE_OPTIONS for two reasons:
#   1. We need to get rid of the init_command that we have for mysql in our
#      default settings.py
#   2. We need to have a high enough timeout for sqlite to obtain a lock.
#      The default is normally 5 seconds but was 0 seconds in some earlier
#      versions of pysqlite. I was getting the 'database is locked' error
#      on testbed-opensuse and found the suggestion of increasing this value
#      here: http://code.djangoproject.com/ticket/9409
#      However, this still doesn't seem to completely solve the issue. It may
#      be an issue with python 2.5's sqlite.
settings.DATABASE_OPTIONS = {'timeout':30}

# Remove the CSRF middleware as our tests don't try to send csrf tokens. We're
# not the only ones who ignore this for testing:
# http://code.djangoproject.com/ticket/11692
new_middleware_classes = list(settings.MIDDLEWARE_CLASSES)
new_middleware_classes.remove('django.contrib.csrf.middleware.CsrfViewMiddleware')
new_middleware_classes.remove('django.contrib.csrf.middleware.CsrfResponseMiddleware')
settings.MIDDLEWARE_CLASSES = tuple(new_middleware_classes)

import django.db

import django.test.utils

from seattlegeni.common.util import log





# Turn off most logging to speed up tests run manually. This can be removed
# to see the plentiful debug output.
log.set_log_level(log.LOG_LEVEL_CRITICAL)





def setup_test_environment():
  """
  Called once before running one or more tests. Must be called before calling
  setup_test_db().
  """
  
  django.test.utils.setup_test_environment()



def teardown_test_environment():
  """
  Called once after running one or more tests. That is, this should be called
  before the test script exits.
  """
  
  django.test.utils.teardown_test_environment()



def setup_test_db():
  
  # Creates a new test database and runs syncdb against it.
  django.db.connection.creation.create_test_db()





def teardown_test_db():
  
  # We aren't going to use any database again in this script, so we give django
  # an empty database name to restore as the value of settings.DATABASE_NAME.
  old_database_name = ''
  django.db.connection.creation.destroy_test_db(old_database_name)
  



"""
<Program>
  action_log_decoratorss.py

<Started>
  Oct 16, 2009

<Author>
  Justin Samuel

<Purpose>
  These define the decorators used to log "actions" to the database.
  
  These are mostly intended to be actions involving vessels such as
  acquiring, releasing, renewing, and expiring vessels (see the models for
  ActionLogEvent and ActionLogVesselDetails for additional information about
  how the data is stored).
"""

import datetime

from seattlegeni.common.api import maindb

from seattlegeni.common.util import decorators

from seattlegeni.website.control import models




def _is_vessel_list(potential_vessel_list):
  """Determine if potential_vessel_list is a list of Vessel objects."""
  
  if not isinstance(potential_vessel_list, list):
    return False

  for i in potential_vessel_list:
    if not isinstance(i, models.Vessel):
      return False
    
  if len(potential_vessel_list) == 0:
    return False
  
  return True





@decorators._simple_decorator
def log_action(func):
  """
  <Purpose>
    Logs information in the database, mostly intended to be actions involving
    vessels (see the models for ActionLogEvent and ActionLogVesselDetails).
  """
  
  # The name "do_logging_func_call" is never seen anywhere but here.
  def do_logging_func_call(*args, **kwargs):
    
    # We are actually going to ignore kwargs and assume keyword arguments
    # aren't being used for the interface calls we are logging with this.
    
    date_started = datetime.datetime.now()
    
    user = None
    second_arg = None
    third_arg = None
    vessel_list = []
    
    # Check if the first arguments is a GeniUser object. We expect it to
    # always be at the moment, so this is just in case things change.
    if args and isinstance(args[0], models.GeniUser):
      user = args[0]
    
    # The interface calls we're using this decorator on may have one or two
    # additional arguments after the geniuser object. If they exist, they
    # are either vessel lists or other values we want to log.
    if len(args) > 1:
      if _is_vessel_list(args[1]):
        vessel_list = args[1]
      else:
        second_arg = str(args[1])
      
    if len(args) > 2:
      if _is_vessel_list(args[2]):
        vessel_list = args[2]
      else:
        third_arg = str(args[2])
    
    try:
      result = func(*args, **kwargs)
      # If a vessel list is returned, that's the one we want even if we took
      # one in as an argument.
      if _is_vessel_list(result):
        vessel_list = result
      was_successful = True
      message = None
      maindb.create_action_log_event(func.__name__, user, second_arg,
                                     third_arg, was_successful, message,
                                     date_started, vessel_list)
      return result
    
    except Exception, e:
      was_successful = False
      message = str(e)
      maindb.create_action_log_event(func.__name__, user, second_arg,
                                     third_arg, was_successful, message,
                                     date_started, vessel_list)
      raise
    
  return do_logging_func_call



"""
<Program>
  transition_onepercentmanyevents_to_canonical.py

<Purpose>
  Recently we have modified our resources file and going 
  to allocate a different amount of resources for each
  of the vessels. In order to do that we are going to have
  to transition the old nodes that are in the onepercentmanyevents
  state to a new state that we are going to call twopercent state.
  This is due to the fact the new vessels will have 2% cpu power
  instead of the 1% power it used to have. It will pass through
  the canonical state to get to the twopercent state. We already
  have a transition script that goes from the canonical state
  to the twopercent state, so we just have to bring all the 
  current nodes that are in the onepercent state to the canonical
  state.

<Author>
  Monzur Muhammad
  monzum@cs.washington.edu

<Usage>
  Ensure that seattlegeni and seattle are in the PYTHONPATH.
  Ensure that the database is setup properly and django settings
    are set correctly.

  python transition_onepercentmanyevents_to_canonical.py
"""

import os

from seattlegeni.common.exceptions import *

from seattlegeni.common.util.decorators import log_function_call

import node_transition_lib



# The full path to the onepercentmanyevents.resources file, including the filename.
RESOURCES_TEMPLATE_FILE_PATH = os.path.join(os.path.dirname(__file__), "resource_files", "onepercentmanyevents.resources")


def main():
  """
  <Purpose>
    This is the main launching point for this script. We are going
    to call the node_transition_lib and start the process of 
    moving from onepercentmanyevents state to the canonical state
    through an intermediate state the moving_to_canonical state.
    This wasy if anything fails, we can go back to the onepercent 
    state.

  <Arguments>
    None.

  <Exceptions>
    None

  <Side Effects>
    A lot of nodes are altered.

  <Return>
    None.
  """


  # Open and read the resource file that is necessary for onepercentmanyevents
  onepercentmanyevents_resource_fd = file(RESOURCES_TEMPLATE_FILE_PATH)
  onepercentmanyevents_resourcetemplate = onepercentmanyevents_resource_fd.read()
  onepercentmanyevents_resource_fd.close()


  # Create a  list that is used to call node_transition_lib.
  # The list will have a list of three tuples. The first 
  # tuple will transition all the onepercentmanyevents state
  # to the moving_to_canonical state. This will use noop functions.
  # We will not actually perform any actions. Next we will transition
  # all the moving_to_canonical state nodes to the canonical state by
  # performing a join operation on all the vessels, to combine all the
  # vessels on one node into one giant node that it was before it got 
  # split. 
  # In the next phase we will transition all the moving_to_canonical
  # state nodes into the onepercentmanyevents state. This is for the 
  # case where we might have failed to combine the vessels and move
  # to canonical state, so we want to restore them back to the 
  # onepercent state.

  
  # We are not going to mark any of the nodes active in the database, 
  # as we are joining the vessels back together.
  mark_nodes_inactive=False

  state_function_arg_tuplelist = [
    ("onepercentmanyevents", "movingto_canonical", 
     node_transition_lib.noop, node_transition_lib.noop, mark_nodes_inactive),

    ("movingto_canonical", "canonical", 
     node_transition_lib.combine_vessels, node_transition_lib.noop, mark_nodes_inactive),

    ("movingto_canonical", "onepercentmanyevents_state", 
     node_transition_lib.split_vessels, node_transition_lib.noop, mark_nodes_inactive, 
     onepercentmanyevents_resourcetemplate)]


  # Some constant variables we will use to call node_transition_lib.
  sleeptime = 10
  process_name = "onepercentmanyevents_to_canonical"
  parallel_instances = 10

  # Call the process_node_and_change_state() function to attempt to 
  # transition all the nodes from onepercentmanyevents to canonical.
  node_transition_lib.process_nodes_and_change_state(state_function_arg_tuplelist, 
                                                     process_name, sleeptime, 
                                                     parallel_instances)






if __name__ == '__main__':
  main()



"""
Example of using the seattleclearinghouse_xmlrpc module's SeattleClearinghouseClient.

This script tries to acquire, renew, and release some vessels and prints out
various information along the way.
"""

# This module must be in your python path (for example, in the same directory
# as the current script is in).
import seattleclearinghouse_xmlrpc


USERNAME = "your_username"

# Only one of either the API key or the private key is needed. If providing
# your API key, be sure to pass the API key to the SeattleClearinghouseClient
# constructor below.
#API_KEY = "your_api_key"
PRIVATE_KEY_STRING = open("your_username.privatekey").read()

# Allowing SSL to be insecure means it will be susceptible to MITM attacks.
# See the instructions in seattleclearinghouse_xmlrpc.py for using SSL securely.
ALLOW_SSL_INSECURE = False


def do_example_acquire_renew_release():
  client = seattleclearinghouse_xmlrpc.SeattleClearinghouseClient(username=USERNAME,
                                                private_key_string=PRIVATE_KEY_STRING,
                                                allow_ssl_insecure=ALLOW_SSL_INSECURE)

  # Obtain general information about the account corresponding to the username
  # and API key.
  account_info = client.get_account_info()
  print("Account info: " + str(account_info))

  # Obtain information on vessels already acquired by this account.
  # The value returned is a list of dictionaries where each dictionary
  # describes a vessel.
  already_acquired_vessels = client.get_resource_info()
  print("Vessels already acquired: " + str(already_acquired_vessels))

  print("Number of vessels this account has available to acquire: " +
        str(account_info['max_vessels'] - len(already_acquired_vessels)))

  # Attempt to acquire two vessels on WAN nodes.
  try:
    # This is a list of dictionaries just like the one above.
    newly_acquired_vessels = client.acquire_wan_resources(2)
  except seattleclearinghouse_xmlrpc.NotEnoughCreditsError, err:
    print("Couldn't acquire resources because we don't have enough vessel " +
          "credits: " + str(err))
    return
  except seattleclearinghouse_xmlrpc.UnableToAcquireResourcesError, err:
    print("Couldn't acquire vessels because SeattleGENI doesn't have enough " +
          "of what we asked for: " + str(err))
    return
  print("Acquired WAN vessel: " + str(newly_acquired_vessels))

  acquired_handle_list = []
  for infodict in newly_acquired_vessels:
    acquired_handle_list.append(infodict['handle'])
 
  # Attempt to renew the two vessels we just acquired. 
  try:
    client.renew_resources(acquired_handle_list)
  except seattleclearinghouse_xmlrpc.NotEnoughCreditsError, err:
    # This is probably not going to happen immediately after acquiring vessels.
    print("Couldn't renew vessels because we are over our vessel credit " +
          "limit: " + str(err))
    return
  print("Renewed the vessel we just acquired.")
  
  print("Summary of all of our acquired vessels:")
  all_vessels_list = client.get_resource_info()
  for vesselinfo in all_vessels_list:
    print("Vessel " + str(vesselinfo["vessel_id"]) +
          " on nodemanager accessible at " + str(vesselinfo["node_ip"]) + ":" +
          str(vesselinfo["node_port"]) + " expires in " +
          str(vesselinfo["expires_in_seconds"]) + " seconds.")

  # Release the two vessels we just acquired.
  client.release_resources(acquired_handle_list)
  print("Released the vessel we just acquired.")



def main():
  try:
    do_example_acquire_renew_release()
  except seattleclearinghouse_xmlrpc.SeattleClearinghouseError:
    # In a real script, you'd want to handle this. Could be an authentication
    # error, communication error, etc. You can use the fine-grained exceptions
    # rather than the base exception of SeattleClearinghouseError to do different
    # things based on what the actual problem was.
    raise


if __name__ == "__main__":
  main()



"""
<Program>
  test_urls.py

<Started>
  8/28/2009

<Author>
  Jason Chen
  jchen@cs.washington.edu

<Purpose>
  Tests that all site urls are being *served* correctly.
  This means that all urls eventually return an HTTP 200 OK, and that 
  non-logged-in, and logged-in users are being redirected to correct pages.

  Note that just because the pages are being served correctly does
  not imply that they *function* correctly. Make sure all other tests
  are run as well, to test the functional correctness of pages.
  
<Notes>
  See test_register.py for an explanation of our usage of the Django test client.
"""

import datetime

# We import the testlib FIRST, as the test db settings 
# need to be set before we import anything else.
from seattlegeni.tests import testlib

from seattlegeni.common.exceptions import *
from seattlegeni.website.control import interface
from seattlegeni.website.control import models
from seattlegeni.website.html import forms

from django.contrib.auth.models import User as DjangoUser

# The django test client emulates a webclient, and returns what django
# considers the final rendered result (in HTML). This allows to test purely the view
# functions. 
from django.test.client import Client



# Declare our mock functions
def mock_get_logged_in_user_throws_DoesNotExistError(request):
  raise DoesNotExistError

def mock_get_logged_in_user(request):
  geniuser = models.GeniUser(username='tester', password='password', email='test@test.com',
                             affiliation='test affil', user_pubkey='user_pubkey',
                             user_privkey='user_privkey', donor_pubkey='donor_pubkey',
                             usable_vessel_port='12345', free_vessel_credits=10)
  return geniuser

def mock_get_total_vessel_credits(geniuser):
  return 10

def mock_get_acquired_vessels(geniuser):
  return ['test', 'test2']

def mock_get_vessel_infodict_list(vessel_list):
  # At a minimum, the expires_in_seconds key must exist because it is used by
  # the myvessels view.
  return [{'expires_in_seconds':1}, 
          {'expires_in_seconds':2}]

def mock_delete_private_key(user):
  pass

# Mock out interface calls
#interface.get_logged_in_user = mock_get_logged_in_user
interface.get_total_vessel_credits = mock_get_total_vessel_credits
interface.get_acquired_vessels = mock_get_acquired_vessels
interface.get_vessel_infodict_list = mock_get_vessel_infodict_list
interface.delete_private_key = mock_delete_private_key

c = Client()



def main():
  
  # Setup test environment
  testlib.setup_test_environment()
  testlib.setup_test_db()
  
  try:
    # Run tests
    get_pages_without_user_logged_in()
    get_pages_with_user_logged_in()
    
    print "All tests passed."
    
  finally:
    testlib.teardown_test_db()
    testlib.teardown_test_environment()



def get_pages_without_user_logged_in():
  # uses the mock get_logged_in_user function that represents no logged in user
  interface.get_logged_in_user = mock_get_logged_in_user_throws_DoesNotExistError
  
  response = c.get('/html/register', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/register.html")
  
  response = c.get('/html/login', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/logout', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/accounts_help', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/help.html")
  
  # The URLs we tell the user to distribute have "download" instead of "html"
  # in them.
  response = c.get('/download/tester/', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "download/installers.html")
  
  # The following pages should all bounce back to the login page,
  # since there is no user logged in currently.
  response = c.get('/html/profile', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/mygeni', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/myvessels', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/help', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/getdonations', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/get_resources', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/del_resource', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/del_all_resources', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/renew_resource', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/renew_all_resources', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  # POST is required for deleting private key.
  response = c.get('/html/del_priv', follow=True)
  assert(response.status_code == 405)
  
  response = c.get('/html/priv_key', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")
  
  response = c.get('/html/pub_key', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")

  response = c.get('/html/change_key', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")

  response = c.get('/html/api_info', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "accounts/login.html")





def get_pages_with_user_logged_in():
  _login_test_user()
  
  response = c.get('/html/register', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/profile.html")
  
  response = c.get('/html/login', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/profile.html")
  
  response = c.get('/html/profile', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/profile.html")
  
  response = c.get('/html/mygeni', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/mygeni.html")
  
  response = c.get('/html/myvessels', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/myvessels.html")
  
  response = c.get('/html/help', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/help.html")
  
  response = c.get('/html/getdonations', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/getdonations.html")
  
  ##########################################################################
  # the following pages expect POST requests. when they see a GET request, #
  # they should bounce the user back to the 'My Vessels' page              #
  ##########################################################################
  response = c.get('/html/get_resources', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/myvessels.html")
  
  response = c.get('/html/del_resource', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/myvessels.html")
  
  response = c.get('/html/del_all_resources', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/myvessels.html")
  
  response = c.get('/html/renew_resource', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/myvessels.html")
  
  response = c.get('/html/renew_all_resources', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/myvessels.html")
  
  response = c.get('/html/api_info', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/api_info.html")
  assert("Your API key is" in response.content)
  
  ##########################################################################
  
  # The change_key page accepts either GET or POST. If it is a GET request,
  # the the forms for changing keys are displayed. If it is a POST and the
  # POST is valid (either a request to generate new keys or upload of a key),
  # the user is shown the profile. If it is an invalid POST, then the same
  # page is redisplayed with a message.
  
  response = c.get('/html/change_key', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/change_key.html")

  # Empty post data is invalid.
  response = c.post('/html/change_key', {}, follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/change_key.html")
  
  ##########################################################################
  
  # POST is required for deleting private key.
  response = c.get('/html/del_priv', follow=True)
  assert(response.status_code == 405)
  
  response = c.post('/html/del_priv', follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/profile.html")
  assert("Your private key has been deleted" in response.content)
  
  # Make sure that priv_key returns an attachment response
  response = c.get('/html/priv_key', follow=True)
  assert(response.status_code == 200)
  assert("attachment" in response['Content-Disposition'])
  assert(".privatekey" in response['Content-Disposition'])
  
  # Make sure that pub_key returns an attachment response
  response = c.get('/html/pub_key', follow=True)
  assert(response.status_code == 200)
  assert("attachment" in response['Content-Disposition'])
  assert(".publickey" in response['Content-Disposition'])

  # Logging out should return an HttpResponseRedirect to login
  response = c.get('/html/logout', follow=True)
  assert(response.status_code == 302)
  assert("/html/login" in response['Location'])





def _login_test_user():
  """
  <Purpose>
    Creates a test user in the test db, and uses the test client to 'login',
    so all views that expect @login_required will now pass the login check. 
  """
  # uses the mock get_logged_in_user function that represents a logged in user
  interface.get_logged_in_user = mock_get_logged_in_user
  
  user = DjangoUser.objects.create_user('tester', 'test@test.com', 'testpassword')
  user.save()
  c.login(username='tester', password='testpassword')
  

if __name__=="__main__":
  main()


"""
Utility functions for the test scripts in this directory/package.
"""

# The seattlegeni testlib must be imported first.
from seattlegeni.tests import testlib

from seattlegeni.common.api import maindb





next_nodeid_number = 0

def create_node_and_vessels_with_one_port_each(ip, portlist, is_active=True):
  
  global next_nodeid_number
  next_nodeid_number += 1
  
  nodeid = "node" + str(next_nodeid_number)
  port = 1234
  version = "10.0test"
  owner_pubkey = "1 2"
  extra_vessel_name = "v1"
  
  node = maindb.create_node(nodeid, ip, port, version, is_active, owner_pubkey, extra_vessel_name)

  single_vessel_number = 2

  for vesselport in portlist:
    single_vessel_name = "v" + str(single_vessel_number)
    single_vessel_number += 1
    vessel = maindb.create_vessel(node, single_vessel_name)
    maindb.set_vessel_ports(vessel, [vesselport])
  
  return node





def create_nodes_on_same_subnet(count, portlist_for_vessels_on_each_node, ip_prefix="127.0.0."):
  # Create 'count' nodes on the same subnet and on each node create a vessel
  # with a single port for each port in 'portlist_for_vessels_on_each_node'.
  for i in range(count):
    ip = ip_prefix + str(i)
    create_node_and_vessels_with_one_port_each(ip, portlist_for_vessels_on_each_node)





def create_nodes_on_different_subnets(count, portlist_for_vessels_on_each_node):
  # Create 'count' nodes on different subnets and on each node create a vessel
  # with a single port for each port in 'portlist_for_vessels_on_each_node'.
  ip_prefix = "127.1."
  ip_suffix = ".0"
  for i in range(count):
    ip = ip_prefix + str(i) + ip_suffix
    create_node_and_vessels_with_one_port_each(ip, portlist_for_vessels_on_each_node)





def create_nat_nodes(count, portlist_for_vessels_on_each_node):
  # Create 'count' nat nodes and on each node create a vessel
  # with a single port for each port in 'portlist_for_vessels_on_each_node'.
  ip_prefix = maindb.NAT_STRING_PREFIX
  for i in range(count):
    ip = ip_prefix + str(i)
    create_node_and_vessels_with_one_port_each(ip, portlist_for_vessels_on_each_node)




"""
<Program>
  test_change_key.py

<Started>
  Oct 13, 2009

<Author>
  Justin Samuel

<Purpose>
  Tests that the change_key view function  handles normal operation and
  exceptions correctly.
  
<Notes>
  See test_register.py for an explanation of our usage of the Django test client.
"""

# We import the testlib FIRST, as the test db settings 
# need to be set before we import anything else.
from seattlegeni.tests import testlib

import tempfile

from seattlegeni.common.exceptions import *
from seattlegeni.website.control import interface
from seattlegeni.website.control import models

from django.contrib.auth.models import User as DjangoUser

# The django test client emulates a webclient, and returns what django
# considers the final rendered result (in HTML). This allows to test purely the view
# functions. 
from django.test.client import Client






# Declare our mock functions
def mock_get_logged_in_user(request):
  geniuser = models.GeniUser(username='tester', password='password', email='test@test.com',
                             affiliation='test affil', user_pubkey='user_pubkey',
                             user_privkey='user_privkey', donor_pubkey='donor_pubkey',
                             usable_vessel_port='12345', free_vessel_credits=10)
  return geniuser





def mock_noop(*args, **kwargs):
  pass
  
  
  
  

c = Client()

interface.change_user_keys = mock_noop





def main():
  
  # Setup test environment
  testlib.setup_test_environment()
  testlib.setup_test_db()
  
  try:
    login_test_user()
    
    test_regenerate_key()
    test_upload_key_no_file()
    test_upload_key_file()
    test_upload_key_file_empty_file()
    test_upload_key_file_invalid_key()
    
    print "All tests passed."
    
  finally:
    testlib.teardown_test_db()
    testlib.teardown_test_environment()




def test_regenerate_key():
  """
  <Purpose>
    Tests submission of form to generate a new user key.
  """
  post_data = {'generate':'yes'}
  response = c.post('/html/change_key', post_data, follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/profile.html")
  assert("Your new keys have been generated" in response.content)





def test_upload_key_no_file():
  """
  <Purpose>
    Tests submitting the upload key form with no key.
  """
  post_data = {}
  response = c.post('/html/change_key', post_data, follow=True)
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/change_key.html")
  assert("You didn&#39;t select a public key file to upload" in response.content)





def test_upload_key_file():
  """
  <Purpose>
    Tests submitting the upload key form with no key.
  """
  uploaded_file = tempfile.NamedTemporaryFile(mode="w+")
  uploaded_file.write('1 2')
  uploaded_file.flush()
  uploaded_file.seek(0)

  post_data = {'pubkey':uploaded_file}
  response = c.post('/html/change_key', post_data, follow=True)
  
  uploaded_file.close()
  
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/profile.html")
  assert("Your public key has been successfully changed" in response.content)




def test_upload_key_file_empty_file():
  """
  <Purpose>
    Tests submitting the upload key form with no key.
  """
  uploaded_file = tempfile.NamedTemporaryFile(mode="w+")

  post_data = {'pubkey':uploaded_file}
  response = c.post('/html/change_key', post_data, follow=True)
  
  uploaded_file.close()
  
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/change_key.html")
  assert("Invalid file uploaded" in response.content)





def test_upload_key_file_invalid_key():
  """
  <Purpose>
    Tests submitting the upload key form with no key.
  """
  uploaded_file = tempfile.NamedTemporaryFile(mode="w+")
  uploaded_file.write('abc')
  uploaded_file.flush()
  uploaded_file.seek(0)

  post_data = {'pubkey':uploaded_file}
  response = c.post('/html/change_key', post_data, follow=True)
  
  uploaded_file.close()
  
  assert(response.status_code == 200)
  assert(response.template[0].name == "control/change_key.html")
  assert("Invalid public key uploaded" in response.content)




  
# Creates a test user in the test db, and uses the test client to 'login',
# so all views that expect @login_required will now pass the login check. 
def login_test_user():
  # uses the mock get_logged_in_user function that represents a logged in user
  interface.get_logged_in_user = mock_get_logged_in_user
  
  user = DjangoUser.objects.create_user('tester', 'test@test.com', 'testpassword')
  user.save()
  c.login(username='tester', password='testpassword')



if __name__ == "__main__":
  main()



"""
<Program>
  check_active_db_nodes.py

<Started>
  16 August 2009

<Author>
  Justin Samuel

<Purpose>
  This script runs an infinite loop of checks over all of the active nodes in
  the database. It will result in inactive and broken nodes or vessels needing
  release/cleanup being marked as such in the database.
  
  This is the only place in seattlegeni where we ensure that nodes that have
  gone offline and no longer advertise get marked as inactive in the database.
  
  This script will not directly result in any state-changing nodemanager
  communication (only node information querying). It uses the check_node()
  function in the seattlegeni.common.util.nodestatus module, which means
  that only the database will be directly modified.
"""

import django.core.mail
import django.db

import sys
import time
import traceback

from seattlegeni.common.api import lockserver
from seattlegeni.common.api import maindb

from seattlegeni.common.util import log

from seattlegeni.common.exceptions import *

from seattlegeni.common.util import nodestatus

from seattlegeni.website import settings



# The number of seconds to sleep at the end of each iteration through all
# active nodes in the database. This only applies when running this as a
# script.
SLEEP_SECONDS_BETWEEN_RUNS = 60

# If True, no database changes will be made marking nodes as inactive or broken
# when this is run as a script. This is independent of the readonly argument
# that can be passed to check_node() if check_node() is called directly from
# another module.
READONLY = False





def main():
  """
  This will run an infinite loop of checks over all of the active nodes in the
  database.
  """
  
  lockserver_handle = lockserver.create_lockserver_handle()

  # Always try to release the lockserver handle, though it's probably not
  # very useful in this case.
  try:
    
    while True:
      
      # Catch unexpected exceptions to log/send mail.
      try:
      
        # We shouldn't be running in production with settings.DEBUG = True. 
        # Just in case, though, tell django to reset its list of saved queries
        # each time through the loop.
        if settings.DEBUG:
          django.db.reset_queries()
        
        # Note: although we include broken but active nodes, we don't change
        # the status of broken nodes to be not broken yet if we don't detect
        # any problems. For now, most of the reason we include broken nodes
        # is so that we can tell which broken nodes are still online. This is
        # because it's not as big of a concern to have a broken node that is
        # quickly offline (e.g. broken nodes in development), but having one be
        # online for an extended period of time is a stronger signal of
        # potentially unknown bugs in the seattlegeni or seattle code.
        active_nodes = maindb.get_active_nodes_include_broken()
        log.info("Starting check of " + str(len(active_nodes)) + " active nodes.")
      
        checked_node_count = 0
        
        for node in active_nodes:
          
          checked_node_count += 1
          log.info("Checking node " + str(checked_node_count) + ": " + str(node))
          
          nodestatus.check_node(node, readonly=READONLY, lockserver_handle=lockserver_handle)
          
        # Print summary info.
        log.info("Nodes checked: " + str(checked_node_count))
        nodes_with_problems = nodestatus.get_node_problem_info()
        nodes_with_problems_count = len(nodes_with_problems.keys())
        log.info("Nodes without problems: " + str(checked_node_count - nodes_with_problems_count))
        log.info("Nodes with problems: " + str(nodes_with_problems_count))
        
        # Print information about the database changes made.
        log.info("Number of database actions taken:")
        actionstaken = nodestatus.get_actions_taken()
        for actionname in actionstaken:
          log.info("\t" + actionname + ": " + str(len(actionstaken[actionname])) + 
                   " " + str(actionstaken[actionname]))
    
        nodestatus.reset_collected_data()
        
        log.info("Sleeping for " + str(SLEEP_SECONDS_BETWEEN_RUNS) + " seconds.")
        time.sleep(SLEEP_SECONDS_BETWEEN_RUNS)
  
      except KeyboardInterrupt:
        raise
  
      except:
        message = "Unexpected exception in check_active_db_nodes.py: " + traceback.format_exc()
        log.critical(message)
    
        # Send an email to the addresses listed in settings.ADMINS
        if not settings.DEBUG:
          subject = "Critical SeattleGeni check_active_db_nodes.py error"
          django.core.mail.mail_admins(subject, message)
          
          # Sleep for ten minutes to make sure we don't flood the admins with error
          # report emails.
          time.sleep(600)

  finally:
    lockserver.destroy_lockserver_handle(lockserver_handle)
  
  



if __name__ == "__main__":
  try:
    main()
  except KeyboardInterrupt:
    log.info("Exiting on KeyboardInterrupt.")
    sys.exit(0)



"""
<Program>
  transition_canonical_to_onepercentmanyevents.py

<Purpose>
  The purpose of this program is to transition nodes from the
  canonical state to the onepercentmanyevents state by bypassing
  through the movingtoonepercentmanyevents state.

<Started>
  August 13, 2009

<Author>
  Monzur Muhammad
  monzum@cs.washington.edu

<Usage>
  Ensure that seattlegeni and seattle are in the PYTHONPATH. 
  Ensure that the database is setup properly and django settings
    are set correctly.

  python transition_canonical_to_onepercentmanyevents.py 
"""


import os
import random
import traceback

from seattlegeni.common.api import backend
from seattlegeni.common.api import maindb
from seattlegeni.common.api import nodemanager

from seattlegeni.common.util.decorators import log_function_call

from seattlegeni.common.exceptions import *

from seattlegeni.node_state_transitions import node_transition_lib






# The full path to the onepercentmanyevents.resources file, including the filename.
RESOURCES_TEMPLATE_FILE_PATH = os.path.join(os.path.dirname(__file__), "resource_files", "onepercentmanyevents.resources")






@log_function_call
def onepercentmanyevents_divide (node_string, node_info, database_nodeobject, onepercent_resourcetemplate):
  """
  <Purpose>
    The purpose of this function is to take a node thats in canonical state
    with one vessel, and split it into the 1% vessels so the vessels can
    be acquired by users.

  <Arguments>
    node_string - the name of the node. ip:port or NAT:port

    node_info - a dictionary containing information about the node

    database_nodeobject - a database object for the node
 
    onepercent_resourcetemplate - the file that has information about resources

  <Exceptions>
    NodeError - Error raised if node is not in the right state 

    NodemanagerCommunicationError - raised if we cannot retrieve the usable ports for a node

    NodeProcessError - raised if unable to split vessels properly

    DatabaseError - raised if unable to modify the database    

  <Side Effects>
    Database gets modified.        

  <Return>
    None
  """

  node_transition_lib.log("Beginning onepercentmanyevents_divide on node: "+node_string)

  # Extract the ip/NAT and the port.
  # Note that the first portion of the node might be an ip or a NAT string.
  (ip_or_nat_string, port_num) = node_transition_lib.split_node_string(node_string)
  
  donated_vesselname = database_nodeobject.extra_vessel_name

  # Retrieve the usable ports list for the node and then shuffle
  # the ports so each vessel gets a random subset of the ports
  usable_ports_list = nodemanager.get_vessel_resources(ip_or_nat_string, port_num, donated_vesselname)['usableports']
  node_transition_lib.log("List of usable ports in node: "+node_string+". "+str(usable_ports_list))
  random.shuffle(usable_ports_list)

  #the vessel that we start with
  current_vessel = donated_vesselname
  node_transition_lib.log("Name of starting vessel: "+current_vessel)

  # Keep splittiing the vessel until we run out of resources.  
  # Note that when split_vessel is called the left vessel 
  # has the leftover (extra vessel)and the right vessel has 
  # the vessel with the exact resources.
  while len(usable_ports_list) >= 10:
    desired_resourcedata = get_resource_data(onepercent_resourcetemplate, usable_ports_list)
                           
    #use the first 10 ports so remove them from the list of usable_ports_list
    used_ports_list = usable_ports_list[:10]
    usable_ports_list = usable_ports_list[10:]

    node_transition_lib.log("Ports we are going to use for the new vessel: "+str(used_ports_list))
    node_transition_lib.log("Starting to split vessel: "+current_vessel)

    # Split the current vessel. The exact vessel is the right vessel 
    # and the extra vessel is the left vessel. 
    try:
      leftover_vessel, new_vessel = backend.split_vessel(database_nodeobject, current_vessel, desired_resourcedata)
    except NodemanagerCommunicationError, e:
      # The object 'e' will already include traceback info that has the actual node error.
      # If the failure is due to inability to split further, that's ok.
      if 'Insufficient quantity:' in str(e):
        node_transition_lib.log("Could not split " + current_vessel + 
                                " any further due to insufficient resource/quantity. " + str(e))
        # We must break out of the while loop here. If we let the exception get,
        # raised, it will look like the transition failed.
        break
      raise

    node_transition_lib.log("Successfully split vessel: "+current_vessel+" into vessels: "+leftover_vessel+" and "+new_vessel)
    current_vessel = leftover_vessel

    # Make sure to update the database and record the new
    # name of the extra vessel as when backend.split_vessels()
    # is called, the old vessel does not exist anymore. 
    # Instead two new vessels are created, where the first
    # vessel is the extra vessel with leftover resources
    # and the second vessel has the actual amount of resources
    maindb.set_node_extra_vessel_name(database_nodeobject, current_vessel)

    #set the user_list for the new vesel to be empty. Remember that user_list is what determines
    #the transition state, and only the extra vessel should have this set.
    backend.set_vessel_user_keylist(database_nodeobject, new_vessel, [])
    node_transition_lib.log("Changed the userkeys for the vessel "+new_vessel+" to []")

    # Add the newly created vessel to the database and then add the ports associated with
    # the vessel to the database also.    
    try:
      node_transition_lib.log("Creating a vessel record in the database for vessel "+new_vessel+" for node "+node_string)
      vessel_object = maindb.create_vessel(database_nodeobject, new_vessel)
      node_transition_lib.log("Setting the vessel ports in the database for vessel "+new_vessel+" with port list: "+str(used_ports_list))
      maindb.set_vessel_ports(vessel_object, used_ports_list)
    except:
      raise node_transition_lib.DatabaseError("Failed to create vessel entry or change vessel entry for vessel: "+
                                               new_vessel+". "+traceback.format_exc())

  # Note: there is one last thing we need to do: set the node as active. We
  # don't want to do this here just in case setting the state key on the node
  # fails. So, rather than add a post-state-key-setting-action for each state,
  # the logic has been put right in processnode() with an if statement to look
  # for the 'onepercentmanyevents_state' as the end state. We could alternately
  # just leave it for the node to be marked as active by the 1pct-to-1pct
  # transition script, but that could introduce a delay of many minutes before
  # the node/donation becomes active.
  node_transition_lib.log("Finished splitting vessels up for the node: "+node_string)


  


@log_function_call
def get_resource_data(onepercent_resourcetemplate, usable_ports_list):
  """
  <Purpose>
    Create the resource_template and return it.

  <Arguments>
    onepercent_resourcetemplate - the resource file 

    usable_ports_list - the list of ports that the node has

  <Exception>
    None

  <Side Effects>
    None

  <Return>
    None
  """

  # Edit the resource file to add the resources in.
  resources_data = onepercent_resourcetemplate % (str(usable_ports_list[0]), str(usable_ports_list[1]), 
                                                  str(usable_ports_list[2]), str(usable_ports_list[3]), 
                                                  str(usable_ports_list[4]), str(usable_ports_list[5]), 
                                                  str(usable_ports_list[6]), str(usable_ports_list[7]),
                                                  str(usable_ports_list[8]), str(usable_ports_list[9]), 
                                                  str(usable_ports_list[0]), str(usable_ports_list[1]), 
                                                  str(usable_ports_list[2]), str(usable_ports_list[3]),
                                                  str(usable_ports_list[4]), str(usable_ports_list[5]), 
                                                  str(usable_ports_list[6]), str(usable_ports_list[7]), 
                                                  str(usable_ports_list[8]), str(usable_ports_list[9]))

  return resources_data





def main():
  """
  <Purpose>
    The main function that calls the process_nodes_and_change_state() function
    in the node_transition_lib passing in the process and error functions.

  <Arguments>
    None
 
  <Exceptions>
    None

  <Side Effects>
    None
  """

  #open and read the resource file that is necessary for onepercentmanyevents
  onepercentmanyevents_resource_fd = file(RESOURCES_TEMPLATE_FILE_PATH)
  onepercentmanyevents_resourcetemplate = onepercentmanyevents_resource_fd.read()
  onepercentmanyevents_resource_fd.close()
  
  """
  build up the tuple list to call process_nodes_and_change_state()
  The transition from canonical to onepercentmanyevents happens in 3 steps.
  Step1: Move the canonical nodes to the movingtoonepercent state (the reason
    this is done is because incase some transition fails, we know that they are 
    going to be in the movingtoonepercent state.
  Step2: Next run the process function and change the state from movingtoonepercent
    state to the onepercentmanyevents state.
  Step3: Find all the nodes that failed to transition from movingtoonepercent
    state to onepercentmanyevents and transition them back to the canonical state.
  """

  state_function_arg_tuplelist = [
    ("canonical", "movingto_onepercentmanyevents", node_transition_lib.noop, node_transition_lib.noop),

    ("movingto_onepercentmanyevents", "onepercentmanyevents", onepercentmanyevents_divide, 
     node_transition_lib.noop,onepercentmanyevents_resourcetemplate),

    ("movingto_onepercentmanyevents", "canonical", node_transition_lib.combine_vessels, 
     node_transition_lib.noop)]
 
  sleeptime = 10
  process_name = "canonical_to_onepercentmanyevents"
  parallel_instances = 10

  #call process_nodes_and_change_state() to start the node state transition
  node_transition_lib.process_nodes_and_change_state(state_function_arg_tuplelist, process_name, sleeptime, parallel_instances) 





if __name__ == '__main__':
  main()



"""
<Program>
  test_onepercentmanyevents_to_onepercentmanyevents.py

<Purpose>
  Test out the onepercentmanyevents_to_onepercentmanyevents transition state. 
  Test to see if the database is updated properly

<Authour>
  Monzur Muhammad
  monzum@cs.washington.edu

<Started>
  Aug 21, 2009
"""

# The seattlegeni testlib must be imported first.
from seattlegeni.tests import testlib

from seattlegeni.node_state_transitions import node_transition_lib
from seattlegeni.node_state_transitions import transition_onepercentmanyevents_to_onepercentmanyevents

from seattlegeni.common.api import maindb

from seattlegeni.node_state_transitions.tests import mockutil



#vessel dictionary for this test
vessels_dict = {}
vessels_dict[mockutil.extra_vessel_name] = {"userkeys" : [node_transition_lib.onepercentmanyeventspublickey],
                                   "ownerkey" : "SeattleGENI",
                                   "ownerinfo" : "",
                                   "status" : "",
                                   "advertise" : True}
vessels_dict["vessel_non_seattlegeni"] = {"userkeys" : ["some random key"],
                                   "ownerkey" : mockutil.donor_key,
                                   "ownerinfo" : "",
                                   "status" : "",
                                   "advertise" : True}
vessels_dict["random_vessel"] = {"userkeys" : ["some random key"],
                                   "ownerkey" : "random key",
                                   "ownerinfo" : "",
                                   "status" : "",
                                   "advertise" : True}








def setup_test():
  """
  <Purpose>
    Prepare everything in order to run the tests.

  <Arguments>
    None

  <Exceptions>
    None

  <Side Efects>
    None

  <Return>
    None
  """

  testlib.setup_test_db()

  # Create a user who has the donation key.
  user_object = maindb.create_user(mockutil.testusername, "password", "example@example.com", "affiliation", 
                    "10 11", "2 2 2", mockutil.donor_key_str)

  # Create a database entry for the node
  node_object = maindb.create_node(mockutil.nodeid_key_str, mockutil.node_ip, mockutil.node_port, "10.0test",
                                  True, mockutil.per_node_key_str, mockutil.extra_vessel_name)

  # Create a donation for user
  maindb.create_donation(node_object, user_object, "Making a donation")
  # Setup all the mock functions
  mockutil.mock_nodemanager_get_node_info(mockutil.nodeid_key, "10.0test", vessels_dict)
  mockutil.mock_lockserver_calls()
  mockutil.mock_backend_generate_key([mockutil.per_node_key_str])
  mockutil.mock_nodemanager_get_vesselresources()
  mockutil.mock_transitionlib_do_advertise_lookup([mockutil.node_address])
  mockutil.mock_backend_set_vessel_owner_key()
  mockutil.mock_backend_split_vessel()
  # set_vessel_user_keylist_call_count won't be called because the
  # node_transition_lib will see that it's a state change to the same state.
  mockutil.mock_backend_set_vessel_user_keylist(None)
 



def run_database_update_test():
  """
  <Purpose>
    Run the test and make sure that the database was modified
    properly, the right keys were set, and generally all information
    is what was expected.

  <Arguments>
    None

  <Exceptions>
    None

  <Side Effects>
    None

  <Return>
    None
  """

  transitionlist = []

  active_nodes_list = maindb.get_active_nodes()
  active_nodes_list[0].is_active = False
  active_nodes_list[0].save()

  transitionlist.append((("startstatename", node_transition_lib.onepercentmanyeventspublickey),
                        ("endstatename", node_transition_lib.onepercentmanyeventspublickey),
                         transition_onepercentmanyevents_to_onepercentmanyevents.update_database,
                         node_transition_lib.noop,
                         transition_onepercentmanyevents_to_onepercentmanyevents.update_database_node))

  (success_count, failure_count) = node_transition_lib.do_one_processnode_run(transitionlist, "startstatename", 1)[0]

  assert(success_count == 1)
  assert(failure_count == 0)

  assert_database_info()

  assert(mockutil.set_vessel_owner_key_call_count == 0)
  # set_vessel_user_keylist_call_count won't be called because the
  # node_transition_lib will see that it's a state change to the same state.
  assert(mockutil.set_vessel_user_keylist_call_count == 0)  





def assert_database_info():

  active_nodes_list = maindb.get_active_nodes()

  assert(len(active_nodes_list) == 1)
  assert(active_nodes_list[0].node_identifier == mockutil.nodeid_key_str)
  assert(active_nodes_list[0].last_known_ip == mockutil.node_ip)
  assert(active_nodes_list[0].last_known_port == mockutil.node_port)
  assert(active_nodes_list[0].extra_vessel_name == mockutil.extra_vessel_name)
  assert(active_nodes_list[0].owner_pubkey == mockutil.per_node_key_str)

  testuser = maindb.get_user(mockutil.testusername)
  donations_list = maindb.get_donations_by_user(testuser)

  assert(len(donations_list) == 1)
  assert(donations_list[0].node == active_nodes_list[0])





def teardown_test():

  # Cleanup the test database.
  testlib.teardown_test_db()




if __name__ == "__main__":
  setup_test()
  try:
    run_database_update_test()
  finally:
    teardown_test()




# The seattlegeni testlib must be imported first.
from seattlegeni.tests import testlib

from seattlegeni.common.api import maindb

from seattlegeni.common.exceptions import *

import unittest





next_nodeid_number = 0

def create_node_and_vessels_with_one_port_each(ip, portlist):
  
  global next_nodeid_number
  next_nodeid_number += 1
  
  nodeid = "node" + str(next_nodeid_number)
  port = 1234
  version = "10.0test"
  is_active = True
  owner_pubkey = "1 2"
  extra_vessel_name = "v1"
  
  node = maindb.create_node(nodeid, ip, port, version, is_active, owner_pubkey, extra_vessel_name)

  single_vessel_number = 2

  for vesselport in portlist:
    single_vessel_name = "v" + str(single_vessel_number)
    single_vessel_number += 1
    vessel = maindb.create_vessel(node, single_vessel_name)
    maindb.set_vessel_ports(vessel, [vesselport])
  
  return node





def create_nodes_on_same_subnet(count, portlist_for_vessels_on_each_node):
  # Create 'count' nodes on the same subnet and on each node create a vessel
  # with a single port for each port in 'portlist_for_vessels_on_each_node'.
  ip_prefix = "ip = 127.0.0."
  for i in range(count):
    ip = ip_prefix + str(i)
    create_node_and_vessels_with_one_port_each(ip, portlist_for_vessels_on_each_node)





def create_nodes_on_different_subnets(count, portlist_for_vessels_on_each_node):
  # Create 'count' nodes on different subnets and on each node create a vessel
  # with a single port for each port in 'portlist_for_vessels_on_each_node'.
  ip_prefix = "127.1."
  ip_suffix = ".0"
  for i in range(count):
    ip = ip_prefix + str(i) + ip_suffix
    create_node_and_vessels_with_one_port_each(ip, portlist_for_vessels_on_each_node)





def create_nat_nodes(count, portlist_for_vessels_on_each_node):
  # Create 'count' nat nodes and on each node create a vessel
  # with a single port for each port in 'portlist_for_vessels_on_each_node'.
  ip_prefix = maindb.NAT_STRING_PREFIX
  for i in range(count):
    ip = ip_prefix + str(i)
    create_node_and_vessels_with_one_port_each(ip, portlist_for_vessels_on_each_node)





def _get_queryset_include_nat(port): 
  """
  Give a nicer though less accurate name to the function as we're using it a
  lot here in the tests.
  """
  return maindb._get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes(port)





def _get_queryset_exclude_nat(port):
  """
  Give a nicer though less accurate name to the function as we're using it a
  lot here in the tests.
  """
  return maindb._get_queryset_of_all_available_vessels_for_a_port_exclude_nat_nodes(port)





def _get_queryset_only_nat(port): 
  """
  Give a nicer though less accurate name to the function as we're using it a
  lot here in the tests.
  """
  return maindb._get_queryset_of_all_available_vessels_for_a_port_only_nat_nodes(port)





class SeattleGeniTestCase(unittest.TestCase):


  def setUp(self):
    # Setup a fresh database for each test.
    testlib.setup_test_db()



  def tearDown(self):
    # Cleanup the test database.
    testlib.teardown_test_db()


  def test_get_queryset_1(self):
    """
    This is ultimately testing
    maindb._get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes()
    """
    
    userport = 100
    
    # Make sure the queryset is initially empty.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(0, queryset.count())

    # Create a node that has no vessels.
    ip = "127.0.0.1"
    portlist = []
    create_node_and_vessels_with_one_port_each(ip, portlist)
    
    # Make sure the queryset is still empty.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(0, queryset.count())
    
    # Create a node that has one vessel, but not on this user's port.
    ip = "127.0.0.2"
    portlist = [userport + 1]
    create_node_and_vessels_with_one_port_each(ip, portlist)
    
    # Make sure the queryset is still empty.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(0, queryset.count())
    
    # Create a node that has one vessel that is on the user's port
    ip = "127.0.0.3"
    portlist = [userport]
    create_node_and_vessels_with_one_port_each(ip, portlist)
    
    # We expect one available vessel.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(1, queryset.count())
    
    # Make sure the vessel is on the last node we created.
    availablevessel = list(queryset)[0]
    self.assertEqual("127.0.0.3", availablevessel.node.last_known_ip)
    
    
    
  def test_get_queryset_node_is_active_changes(self):
    """
    This is ultimately testing
    maindb._get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes()
    """
    
    userport = 100
    
    # Create a node that has three vessels but only one vessel on the user's port.
    ip = "127.0.0.1"
    portlist = [userport - 1, userport, userport + 1]
    node = create_node_and_vessels_with_one_port_each(ip, portlist)
    
    # We expect one available vessel.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(1, queryset.count())
    
    # Now update the node to mark it as inactive.
    node.is_active = False
    node.save()
    
    # We expect zero available vessels.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(0, queryset.count())
    
    # Now update the node to mark it as active again.
    node.is_active = True
    node.save()
    
    # We expect one available vessel.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(1, queryset.count())



  def test_get_queryset_node_is_broken_changes(self):
    """
    This is ultimately testing
    maindb._get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes()
    """
    
    userport = 100
    
    # Create a node that has three vessels but only one vessel on the user's port.
    ip = "127.0.0.1"
    portlist = [userport - 1, userport, userport + 1]
    node = create_node_and_vessels_with_one_port_each(ip, portlist)
    
    # We expect one available vessel.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(1, queryset.count())
    
    # Now update the node to mark it as inactive.
    node.is_broken = True
    node.save()
    
    # We expect zero available vessels.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(0, queryset.count())
    
    # Now update the node to mark it as active again.
    node.is_broken = False
    node.save()
    
    # We expect one available vessel.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(1, queryset.count())


    
  def test_get_queryset_vessel_is_dirty_changes(self):
    """
    This is ultimately testing
    maindb._get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes()
    """
    
    userport = 100
    
    # Create two nodes that have three vessels but only one vessel on the user's port.
    portlist = [userport - 1, userport, userport + 1]
    ip = "127.0.0.1"
    create_node_and_vessels_with_one_port_each(ip, portlist)
    ip = "127.0.0.2"
    create_node_and_vessels_with_one_port_each(ip, portlist)

    # We expect two available vessels.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(2, queryset.count())
    
    # Now mark one of the two available vessels as dirty.
    vessel = queryset[0]
    vessel.is_dirty = True
    vessel.save()
    
    # We expect one available vessel, and it shouldn't be the dirty one.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(1, queryset.count())
    self.assertNotEqual(vessel, queryset[0])
    
    

  def test_get_queryset_vessel_acquired_by_user_changes(self):
    """
    This is ultimately testing
    maindb._get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes()
    """
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    
    userport = user.usable_vessel_port
    
    # Create two nodes that have three vessels but only one vessel on the user's port.
    portlist = [userport - 1, userport, userport + 1]
    ip = "127.0.0.1"
    create_node_and_vessels_with_one_port_each(ip, portlist)
    ip = "127.0.0.2"
    create_node_and_vessels_with_one_port_each(ip, portlist)

    # We expect two available vessels.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(2, queryset.count())
    
    # Mark one of the vessels as acquired.
    vessel = queryset[0]
    maindb.record_acquired_vessel(user, vessel)
    
    # We expect one available vessel, and it shouldn't be the acquired one.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(1, queryset.count())
    self.assertNotEqual(vessel, queryset[0])
    
    # Release the vessel. It should still be dirty.
    maindb.record_released_vessel(vessel)
    
    # We expect one available vessel, and it shouldn't be the acquired one.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(1, queryset.count())
    self.assertNotEqual(vessel, queryset[0])
    
    # Mark the vessel as clean (as if the backend cleaned it up).
    maindb.mark_vessel_as_clean(vessel)
    
    # We expect two available vessels.
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(2, queryset.count())
    

    
# TODO: add a test for broken nodes if/when an is_broken flag is added to nodes



  def test_get_queryset_different_node_types(self):
    
    userport = 12345
    
    # We choose the numbers of each type of node in a way that helps ensure
    # that we don't accidentally pass the test if something is going wrong.
    
    # We will get one vessel on each created node for each port in portlist
    # and there will be only that one port on the vessel.
    portlist = [userport - 1, userport, userport + 1]
    
    create_nodes_on_same_subnet(3, portlist)
    create_nodes_on_different_subnets(7, portlist)
    create_nat_nodes(13, portlist)

    # We expect 23 available total vessels (that is, one vessel on each node).
    queryset = _get_queryset_include_nat(userport)
    self.assertEqual(23, queryset.count())
    
    # We expect 10 available vessels not on nat nodes.
    queryset = _get_queryset_exclude_nat(userport)
    self.assertEqual(10, queryset.count())
    
    # We expect 13 available vessels on nat nodes.
    queryset = _get_queryset_only_nat(userport)
    self.assertEqual(13, queryset.count())




  def test_get_available_rand_vessels(self):
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    
    userport = user.usable_vessel_port
    
    # We choose the numbers of each type of node in a way that helps ensure
    # that we don't accidentally pass the test if something is going wrong.
    
    # We will get one vessel on each created node for each port in portlist
    # and there will be only that one port on the vessel.
    portlist = [userport - 1, userport, userport + 1]
    
    create_nodes_on_same_subnet(3, portlist)
    create_nodes_on_different_subnets(7, portlist)
    create_nat_nodes(13, portlist)
      
    # Request 0 vessels, make sure it raises a AssertionError.
    self.assertRaises(AssertionError, maindb.get_available_rand_vessels, user, 0)
    
    # Request a negative number of vessels, make sure it raises a AssertionError.
    self.assertRaises(AssertionError, maindb.get_available_rand_vessels, user, -1)
    
    # We expect there to be 23 available rand vessels (one vessel on each node
    # including the nat nodes).
    
    # Request 1 vessel, make sure we get back more than 1 potential vessels.
    vessel_list = maindb.get_available_rand_vessels(user, 1)
    self.assertTrue(len(vessel_list) > 1)
      
    # Request 5 vessels, make sure we get back more than 5 potential vessels.
    vessel_list = maindb.get_available_rand_vessels(user, 5)
    self.assertTrue(len(vessel_list) > 5)
      
    # Request 23 vessels, make sure we get back all 23 vessels we expect.
    vessel_list = maindb.get_available_rand_vessels(user, 23)
    self.assertEqual(23, len(vessel_list))
    
    # Request 24 vessels, make sure we get an exception.
    self.assertRaises(UnableToAcquireResourcesError, maindb.get_available_rand_vessels, user, 24)

      
      
  def test_get_available_wan_vessels(self):
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    
    userport = user.usable_vessel_port
    
    # We choose the numbers of each type of node in a way that helps ensure
    # that we don't accidentally pass the test if something is going wrong.
    
    # We will get one vessel on each created node for each port in portlist
    # and there will be only that one port on the vessel.
    portlist = [userport - 1, userport, userport + 1]
    
    create_nodes_on_same_subnet(3, portlist)
    create_nodes_on_different_subnets(7, portlist)
    create_nat_nodes(13, portlist)
      
    # Request 0 vessels, make sure it raises a AssertionError.
    self.assertRaises(AssertionError, maindb.get_available_wan_vessels, user, 0)
    
    # Request a negative number of vessels, make sure it raises a AssertionError.
    self.assertRaises(AssertionError, maindb.get_available_wan_vessels, user, -1)
    
    # We expect there to be 8 available rand vessels (one vessel on each node
    # non-nat node on each subnet, and there are 8 subnets).
    
    # Request 1 vessel, make sure we get back more than 1 potential vessels.
    vessel_list = maindb.get_available_wan_vessels(user, 1)
    self.assertTrue(len(vessel_list) > 1)
      
    # Request 5 vessels, make sure we get back more than 5 potential vessels.
    vessel_list = maindb.get_available_wan_vessels(user, 5)
    self.assertTrue(len(vessel_list) > 5)
      
    # Request 8 vessels, make sure we get back all 8 vessels we expect.
    vessel_list = maindb.get_available_wan_vessels(user, 8)
    self.assertEqual(8, len(vessel_list))
    
    # Request 9 vessels, make sure we get an exception.
    self.assertRaises(UnableToAcquireResourcesError, maindb.get_available_wan_vessels, user, 9)

      
      
  def test_get_available_nat_vessels(self):
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    
    userport = user.usable_vessel_port
    
    # We choose the numbers of each type of node in a way that helps ensure
    # that we don't accidentally pass the test if something is going wrong.
    
    # We will get one vessel on each created node for each port in portlist
    # and there will be only that one port on the vessel.
    portlist = [userport - 1, userport, userport + 1]
    
    create_nodes_on_same_subnet(3, portlist)
    create_nodes_on_different_subnets(7, portlist)
    create_nat_nodes(13, portlist)
      
    # Request 0 vessels, make sure it raises a AssertionError.
    self.assertRaises(AssertionError, maindb.get_available_nat_vessels, user, 0)
    
    # Request a negative number of vessels, make sure it raises a AssertionError.
    self.assertRaises(AssertionError, maindb.get_available_nat_vessels, user, -1)
    
    # We expect there to be 13 available nat vessels (one vessel on each node
    # nat node).
    
    # Request 1 vessel, make sure we get back more than 1 potential vessels.
    vessel_list = maindb.get_available_nat_vessels(user, 1)
    self.assertTrue(len(vessel_list) > 1)
      
    # Request 5 vessels, make sure we get back more than 5 potential vessels.
    vessel_list = maindb.get_available_nat_vessels(user, 5)
    self.assertTrue(len(vessel_list) > 5)
      
    # Request 13 vessels, make sure we get back all 13 vessels we expect.
    vessel_list = maindb.get_available_nat_vessels(user, 13)
    self.assertEqual(13, len(vessel_list))
    
    # Request 14 vessels, make sure we get an exception.
    self.assertRaises(UnableToAcquireResourcesError, maindb.get_available_nat_vessels, user, 14)



  def test_get_available_lan_vessels_by_subnet(self):
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    
    userport = user.usable_vessel_port
    
    # We choose the numbers of each type of node in a way that helps ensure
    # that we don't accidentally pass the test if something is going wrong.
    
    # We will get one vessel on each created node for each port in portlist
    # and there will be only that one port on the vessel.
    portlist = [userport - 1, userport, userport + 1]
    
    create_nodes_on_same_subnet(29, portlist)
    create_nodes_on_different_subnets(7, portlist)
    create_nat_nodes(13, portlist)
    
    # Request 0 vessels, make sure it raises a AssertionError.
    self.assertRaises(AssertionError, maindb.get_available_lan_vessels_by_subnet, user, 0)
    
    # Request a negative number of vessels, make sure it raises a AssertionError.
    self.assertRaises(AssertionError, maindb.get_available_lan_vessels_by_subnet, user, -1)
    
    # We expect there to be 7 subnets with a single available vessel for the
    # user (the 7 on differnt subnets created above) and 1 subnet with 29
    # available vessels for the user (the 29 nodes on the same subnet created
    # above).
    
    # Request 1 vessel, make sure we get back a list of 8 subnets where one
    # subnet has more than one available vessel and the other 7 have only
    # one available vessel.
    subnet_vessel_list = maindb.get_available_lan_vessels_by_subnet(user, 1)
    self.assertEqual(8, len(subnet_vessel_list))
    
    vessel_list_sizes = []
    for vessel_list in subnet_vessel_list:
      vessel_list_sizes.append(len(vessel_list))
    vessel_list_sizes.sort()
    
    self.assertEqual([1, 1, 1, 1, 1, 1, 1], vessel_list_sizes[:7])
    self.assertTrue(vessel_list_sizes[7] > 1)
      
    # Request 5 vessels, make sure we get back a list that has one subnet and
    # in that subnet is a list of more than 5 potential vessels.
    subnet_vessel_list = maindb.get_available_lan_vessels_by_subnet(user, 5)
    self.assertEqual(1, len(subnet_vessel_list))
    self.assertTrue(len(subnet_vessel_list[0]) > 5)
      
    # Request 29 vessels, make sure we get back a list that has one subnet and
    # in that subnet is a list of 29 potential vessels.
    subnet_vessel_list = maindb.get_available_lan_vessels_by_subnet(user, 29)
    self.assertEqual(1, len(subnet_vessel_list))
    self.assertEqual(29, len(subnet_vessel_list[0]))
    
    # Request 30 vessels, make sure we get an exception.
    self.assertRaises(UnableToAcquireResourcesError, maindb.get_available_lan_vessels_by_subnet, user, 30)





def run_test():
  unittest.main()



if __name__ == "__main__":
  run_test()
